import pygst
pygst.require('0.10')
import gst
from xl import common, xlmisc
from mutagen.mp3 import HeaderNotFoundError
import os.path, gobject, re

__all__ = ['flac', 'mp3', 'm4a', 'ogg', 'wma', 'mpc', 'wv', 'tta']

from xl.media import flac, mp3, mp4, mpc, ogg, tta, wav, wma, wv

formats = {
    'aac': mp4,
    'ac3': None,
    'flac': flac,
    'm4a': mp4,
    'mp+': mpc,
    'mp2': mp3,
    'mp3': mp3,
    'mp4': mp4,
    'mod': None,
    'mpc': mpc,
    'oga': ogg,
    'ogg': ogg,
    's3m': None,
    'tta': tta,
    'wav': wav,
    'wma': wma,
    'wv': wv,
    'xm': None,
}

SUPPORTED_MEDIA = ['.' + ext for ext in formats.iterkeys()]

# Generic functions

def write_tag(tr):
    """
        Writes a tag
    """
    (path, ext) = os.path.splitext(tr.loc.lower())
    ext = ext[1:]

    if not formats.get(ext):
        xlmisc.log("Writing metadata to type '%s' is not supported" % ext)
        return

    formats[ext].write_tag(tr)

class timetype(long):
    """
        I am just extending long so that we know when to convert a long to a
        time when displaying the tracklist. I don't just send the trackslist
        the time in the 00:00 format because it won't sort correctly (I want
        it to sort numerically instead of alphabetically.
    """
    def __init__(self, num=None):
        """
            Initializes the class
        """
        long.__init__(self, num)
        self.stream = False

class Track(gobject.GObject): 
    """
        Represents a generic single track.
    """
    type = 'track'
    def __init__(self, *args, **kwargs):
        """
            Loads and initializes the tag information
            Expects the path to the track as an argument
        """
        gobject.GObject.__init__(self)
        
        self.tags = common.ldict()

        self.time_played = 0
        self.read_from_db = False
        self.blacklisted = 0
        self.next_func = None
        self.start_time = 0

        self.set_info(*args, **kwargs)

        try:
            self.ext = os.path.splitext(self.loc.lower())[1]
            self.ext = self.ext.replace('.', '')
        except:
            self.ext = None 


    def full_status(self, player):
        """
            Returns a string representing the status of the current track
        """
        status = "playing"
        if player.is_paused(): status = "paused"

        value = player.get_current_position()
        duration = self.duration * gst.SECOND

        if duration == -1:
            real = 0
        else:
            real = value * duration / 100
        seconds = real / gst.SECOND

        return "status: %s self: %s artist: %s " \
            "album: %s length: %s position: %%%d [%d:%02d]" % (status,
                self.title,
                self.artist, self.album, self.length,
                value, seconds / 60, seconds % 60)

    def is_file(self):
        return self.type == "file"

    def is_multi(self):
        """
            Does the track support multiple tag values?
        """
        return formats[self.ext].is_multi()

    def set_info(self,loc="", title="", artist="",  
        album="", disc_id=0, genre="",
        track=0, length=0, bitrate=0, year="", 
        modified=0, user_rating=0, rating=0, blacklisted=0, time_added='', 
        encoding=xlmisc.get_default_encoding(), playcount=0):
    
    
        """
            Sets track information
        """
        
        # Doesn't matter what charset we use here, as long as we use
        # the same one when we decode (or encode as it were)
        if type(loc) is unicode:
            self._loc = loc
        else:        
            try:
                self._loc = unicode(loc, xlmisc.get_default_encoding())
            except (UnicodeDecodeError, TypeError):
                self._loc = loc

        self._encoding = encoding
        self._bitrate = bitrate
        self._len = length
        self.connections = []
        self.date = year
        self.playing = 0
        self.submitted = False
        self.last_position = 0
        self.bitrate = bitrate
        self.modified = modified
        self.blacklisted = blacklisted
        self._rating = user_rating
        self.system_rating = rating
        self.time_added = time_added
        self.playcount = playcount
    
        for tag, val in {'title': title, 'artist': artist, 'album':album,\
                        'genre': genre, 'discnumber':disc_id,\
                        'tracknumber':track}.iteritems():
            self.set_tag(tag, val)

    def can_change(self, tag):
        """
            If the tag in question is supported by the file format, return True
        """
        return formats[self.ext].can_change(tag)

    def write_tag(self, db=None):
        """
            Writes the tag information to the database
        """

        if db:
            mod = os.stat(self.loc).st_mtime
            artist_id = tracks.get_column_id(db, 'artists', 'name',
                self.artist)
            album_id = tracks.get_album_id(db, artist_id, self.album)
            path_id = tracks.get_column_id(db, 'paths', 'name', self.loc)

            db.execute("UPDATE tracks SET title=?, artist=?, " \
                "album=?, disc_id=?, genre=?, year=?, modified=?, track=? WHERE path=?",
                (self.title, artist_id, album_id, self.disc_id, self.genre,
                self.date, mod, self.track, path_id))

    def __str__(self):
        """
            Returns a string representation of the track.

            Note that for now this string is only suitable for logging because
            it is not translated and does not take into account when album or
            artist is empty.
        """
        return "%s from %s by %s" % (self.title, self.album, self.artist)
   
    def get_tag(self, tag):
        """
            Common function for getting a tag.
            Simplifies a list into a single string separated by " / ".
        """
        values = self.tags.get(tag)
        if values:
            values = (common.to_unicode(x, self.encoding) for x in values
                if x not in (None, ''))
            return u" / ".join(values)
        return u""

    def set_tag(self, tag, values, append=False):
        """
            Common function for setting a tag.
            Expects a list (even for a single value)
        """
        if not isinstance(values, list): values = [values]
        # filter out empty values and convert to unicode
        values = (common.to_unicode(x, self.encoding) for x in values
            if x not in (None, ''))
        if append:
            self.tags[tag].extend(values)
        else:
            self.tags[tag] = list(values)

   # ========== Getters and setters ============

    def get_filename(self):
        """
            Returns the base filename of the track location
        """
        return os.path.basename(self.io_loc)

    def set_track(self, t): 
        """
            Sets the track number
        """
        self.set_tag('tracknumber', t)
    
    def get_track(self): 
        """
            attempts to convert the track number to an int, otherwise it
            just returns -1
        """
        if self.type == 'stream':   
            return -1

        t = self.get_tag('tracknumber')
        if type(t) is int: return t

        b = t.find('/')

        if b > -1: t = t[0:b]

        try:
            return int(t)
        except:
            return -1
    
    def get_bitrate(self): 
        """
            Returns the bitrate
        """
        if self.type == 'stream':
            if self._bitrate:
                try:
                    return "%sk" % self._bitrate.replace('k', '')
                except AttributeError:
                    return str(self._bitrate) + "k"
            else:
                return ''
        try:
            rate = int(self._bitrate) / 1000
            if rate: return "%dk" % rate
            else: return ""
        except:
            return self._bitrate
    
    def get_rating(self): 
        """
            Gets the rating
        """
        try:
            return "* " * self._rating
        except TypeError:
            return ""
    
    def set_rating(self, rating): 
        """
            Sets the rating
        """
        self._rating = rating
        self.user_rating = rating

    def get_title(self): 
        """
            Returns the title of the track from the id3 tag
        """
        ret = self.get_tag('title')
        if not ret:
            return os.path.basename(self.loc)
        else:
            return ret
    
    def set_title(self, value): 
        """
            Sets the title
        """
        self.set_tag('title', value)

    def set_artist(self, value):
        """
            Sets the artist
        """
        self.set_tag('artist', value)

    def get_artist(self):
        """
            Gets the artist
        """
        return self.get_tag('artist')

    def set_album(self, value):
        """
            Sets the album
        """
        self.set_tag('album', value)

    def get_album(self):
        """
            Gets the album
        """
        return self.get_tag('album')

    def get_encoding(self):
        """
            Gets the encoding used for the metadata
        """
        return self._encoding

    def set_encoding(self, value):
        """
            Sets the encoding, translating info from the previous one
        """
        for tag in xlmisc.VALID_TAGS:
            try:
                self.tags[tag] = unicode(self.tags[tag].encode(self.encoding), value)
            except AttributeError:
                pass

        self._encoding = value

    def get_len(self): 
        """
            Returns the length of the track in the format minutes:seconds
        """
        sec = int(round(float(self._len)))
        return "%s:%02d" % divmod(sec, 60)
    
    def set_len(self, value): 
        """
            Sets the length
        """
        if value == "": value = 0
        self._len = value

    def get_duration(self): 
        """
            Gets the duration as an integer
        """
        if self._len == '': self._len = 0
        return timetype(self._len)
    
    def set_bitrate(self, rate):
        """
            Gets the bitrate for this track
        """
        self._bitrate = rate

    def get_loc(self):
        """
            Gets the location as unicode (might contain garbled characters)
        """
        return self._loc

    def get_loc_for_io(self):
        """
            Gets the location as ascii. Should always be correct, see set_loc.
        """
        return self._loc.encode(xlmisc.get_default_encoding())

    def set_loc(self, value):
        """
            Sets the location. It is always in unicode.
            If the value is not unicode, convert it into unicode using some
            default mapping. This way, when we want to access the file, we
            decode it back into the ascii and don't worry about botched up
            characters (ie the value should be exactly identical to the one given)
        """
        self._loc = common.to_unicode(value, xlmisc.get_default_encoding())

    def get_disc(self):
        """
            Gets the disc number
        """
        return self.get_tag('discnumber')


    def set_disc(self, value):
        """
            Sets the disc number
        """
        self.set_tag('discnumber', value)

    def get_version(self):
        """
            Get the version (ie "remixed by" etc)
        """
        return self.get_tag('version')

    def set_version(self, value):
        """
            Set the version
        """
        self.set_tag('version', value)

    def get_performer(self):
        """
            Get the lead performer/soloist
        """
        return self.get_tag('performer')

    def set_performer(self, value):
        """
            Set performer
        """
        self.set_tag('performer', value)

    def get_copyright(self):
        """
            Get copyright information
        """
        return self.get_tag('copyright')

    def set_copyright(self, value):
        """
            Set copyright
        """
        self.set_tag('copyright', value)

    def get_publisher(self):
        """
            Get the publisher (ie record label etc)
        """
        return self.get_tag('publisher')

    def set_publisher(self, value):
        """
            Set the publisher
        """
        self.set_tag('publisher', value)

    def get_date(self):
        """
            Get the recording date
        """
        return self.get_tag('date')

    def set_date(self, value):
        """
            Set the date
        """
        # FIXME: check if the value is a valid ISO 8601 date
        self.set_tag('date', value)

    def get_isrc(self):
        """
            Get the isrc (international standard recording code)
        """
        return self.get_tag('isrc')

    def set_isrc(self, value):
        """
            Set the isrc
        """
        self.set_tag('isrc', value)

    def get_genre(self):
        """
            Get the genre
        """
        return self.get_tag('genre')

    def set_genre(self, value):
        """
            Set the genre
        """
        self.set_tag('genre', value)

    bitrate = property(get_bitrate, set_bitrate)
    duration = property(get_duration)
    encoding = property(get_encoding, set_encoding)
    filename = property(get_filename)
    io_loc = property(get_loc_for_io, None)
    length = property(get_len, set_len)
    loc = property(get_loc, set_loc)
    rating = property(get_rating, set_rating)

    # Data written to tags
    album = property(get_album, set_album)
    artist = property(get_artist, set_artist)
    copyright = property(get_copyright, set_copyright)
    date = property(get_date, set_date)
    disc_id = property(get_disc, set_disc)
    genre = property(get_genre, set_genre)
    isrc = property(get_isrc, set_isrc)
    performer = property(get_performer, set_performer)
    publisher = property(get_publisher, set_publisher)
    title = property(get_title, set_title)
    track = property(get_track, set_track)
    version = property(get_version, set_version)
    year = property(get_date, set_date) # backwards compatibility



def read_from_path(uri, track_type=Track):
    """
        Reads tags from a specified uri
    """
    (path, ext) = os.path.splitext(uri.lower())
    ext = ext[1:]

    #if ext not in formats:
    #    xlmisc.log('%s format is not understood' % ext)
    #    return None

    tr = track_type(uri)

    if tr.type != 'device':
        tr.type = 'file' 

    format = formats.get(ext)
    if not format: return tr

    try:
        format.fill_tag_from_path(tr)
    except HeaderNotFoundError:
        print "Possibly corrupt file: " + uri
        return None
    except:
        xlmisc.log_exception()
        return None

    return tr




# Copyright (C) 2007 Aren Olson
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 1, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import pygtk
pygtk.require("2.0")
import gtk
from gettext import gettext as _
import xl.plugins as plugins

PLUGIN_NAME = _("Panel List")
PLUGIN_AUTHORS = ['Aren Olson <reacocard@gmail.com>']
PLUGIN_VERSION = '0.1.3'
PLUGIN_DESCRIPTION = _(r"""Moves the sidepanel's tab list into a drop-down menu.""")

PLUGIN_ENABLED = False
PLUGIN_ICON = None

APP = None
CBOX = None
NOTEBOOK = None
CONNS = plugins.SignalContainer()


def switch_page(widget=None, junk=None):
    global APP, CBOX, CONNS, NOTEBOOK

    NOTEBOOK.set_current_page(CBOX.get_active())

def update_list(spam=None, eggs=None, sausage=None):
    global APP, CBOX, CONNS, NOTEBOOK

    for i in range(20): #bad!!
        CBOX.remove_text(0)
    for pagenum in range(NOTEBOOK.get_n_pages()):
        name = NOTEBOOK.get_tab_label(NOTEBOOK.get_nth_page(pagenum))
        CBOX.append_text(name.get_text())
    change_active()

def change_active(spam=None, eggs=None, sausage=None):
    global APP, CBOX, CONNS, NOTEBOOK

    CBOX.set_active(NOTEBOOK.get_current_page())

def initialize():
    global APP, CBOX, CONNS, NOTEBOOK

    NOTEBOOK = APP.xml.get_widget('side_notebook')

    xml_string = '''<glade-interface>
                    <widget class="GtkComboBox" id="panels_combo_box">
                        <property name="visible">True</property>
                        <property name="items" translatable="yes"></property>
                        <property name="add_tearoffs">False</property>
                        <property name="focus_on_click">True</property>
                    </widget>
                    </glade-interface>'''

    combo_xml = gtk.glade.xml_new_from_buffer(xml_string, len(xml_string))

    CBOX = combo_xml.get_widget('panels_combo_box')

    #CBOX = gtk.ComboBox()
    side_panel = APP.xml.get_widget('side_panel')
    side_panel.pack_start(CBOX, False, False, 2)
    side_panel.reorder_child(CBOX, 0)
    CBOX.show()
    update_list()
    change_active()
    CBOX.connect('changed', switch_page)
    CONNS.connect(NOTEBOOK, 'page-added', update_list)
    CONNS.connect(NOTEBOOK, 'page-removed', update_list)
    CONNS.connect(NOTEBOOK, 'page-reordered', update_list)
    #CONNS.connect(NOTEBOOK, 'switch-page', change_active)
    NOTEBOOK.set_show_tabs(False)

    return True

def destroy():
    global APP, CBOX, CONNS, NOTEBOOK
    CBOX.destroy()
    CBOX = None
    NOTEBOOK.set_show_tabs(True)
    CONNS.disconnect_all()



# Copyright (C) 2007 Mathieu Virbel <tito@bankiz.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 1, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import gtk, threading, gobject, cgi
from gettext import gettext as _
from BaseHTTPServer import BaseHTTPRequestHandler, HTTPServer

PLUGIN_NAME = _('HTTP Server Control')
PLUGIN_AUTHORS = ['Mathieu Virbel <tito@bankiz.org>']
PLUGIN_VERSION = '0.2.2'
PLUGIN_DESCRIPTION = _('Open an HTTP Server to control exaile from a remote host')
PLUGIN_ENABLED = False
button = gtk.Button()
PLUGIN_ICON = button.render_icon('gtk-info', gtk.ICON_SIZE_MENU)
button.destroy()

DEFAULT_PORT = 10000

APP = None
eh_thread = None



# -----------------------------------------------------------------------------
# Utilities
# -----------------------------------------------------------------------------

def tag(tagname, value):
    """
        Construct a xml tag
    """
    return '<%s>%s</%s>' % (tagname, value, tagname)

class TrackInfo:
    """
        Fetch some infos from a Track
        (In case we have no track, set some default value)
        (We also can have reduce information for display a playlist)
    """

    def __init__(self, track = None, playlistmode=False):
        self.infos = {}
        self.infos['filename']	= ''
        self.infos['title']		= _('None')
        self.infos['artist']	= _('None')
        self.infos['album']		= _('None')
        self.infos['bitrate']	= _('None')
        self.infos['rating']	= _('Unknown')
        self.infos['len']		= '0:00'
        self.infos['disc']		= _('None')
        self.infos['genre']		= _('None')
        self.infos['duration']	= '0'
        self.infos['is_paused']	= APP.player.is_paused()
        self.infos['is_playing']= APP.player.is_playing()

        if track is not None:
            self.infos['filename'] = cgi.escape(track.get_filename())
            self.infos['title'] = cgi.escape(track.get_title())
            self.infos['artist'] = cgi.escape(track.get_artist())
            self.infos['album'] = cgi.escape(track.get_album())
            self.infos['bitrate'] = cgi.escape(track.get_bitrate())
            self.infos['rating'] = cgi.escape(track.get_rating())
            self.infos['len'] = cgi.escape(track.get_len())
            self.infos['disc'] = cgi.escape(track.get_disc())
            self.infos['genre'] = cgi.escape(track.get_genre())
            self.infos['duration'] = cgi.escape(str(track.get_duration()))

        if not playlistmode:
            self.infos['position'] = '0'
            if track is not None:
                self.infos['position'] = cgi.escape(str(APP.player.get_current_position()))



# -----------------------------------------------------------------------------
# Callbacks for HTTP Request
# -----------------------------------------------------------------------------

def eh_page_rpc_current(r):
    """
        Get informations from current track
        Response is in XML.
    """
    r.send_response(200, 'OK')
    r.send_header('Pragma', 'no-cache')
    r.send_header('Content-type', 'text/xml; charset=utf-8')
    r.end_headers()

    ti = TrackInfo(APP.player.current)
    r.wfile.write('<?xml version="1.0"?>')
    r.wfile.write('<track>')
    for key in ti.infos:
        r.wfile.write(tag(key, ti.infos[key]))
    r.wfile.write('</track>')


def eh_page_cover_current(r):
    """
        Return the current track cover in a HTTP Response
        Support only jpeg or png file
    """
    track = APP.player.current
    if track is None:
        r.send_error(404,'Not Found')
        return
    filename = APP.cover_manager.fetch_cover(track, 1)
    if filename is None or filename == '':
        r.send_error(404, 'Not Found')
        return

    handle = open(filename, 'rb')
    if handle is None:
        self.send_error(404, 'Not Found')
        return

    r.send_response(200, 'OK')
    r.send_header('Pragma', 'no-cache')
    if filename.rfind('.jpg') > 0:
        r.send_header('Content-type', 'image/jpeg')
    else:
        r.send_header('Content-type', 'image/png')
    r.end_headers()

    data = handle.read()
    handle.close()

    r.wfile.write(data)


def eh_page_file(r):
    """
        Return a static file from zipfile
        Filter is done before.
    """
    r.send_response(200, 'OK')
    r.send_header('Pragma', 'no-cache')
    if r.path.rfind('.css') > 0:
        r.send_header('Content-type', 'text/css')
    else:
        r.send_header('Content-type', 'text/html; charset=utf-8')
    r.end_headers()

    path = r.path
    if path == '/':
        path = '/index.html'

    data = ZIP.get_data('data%s' % path)
    r.wfile.write(data)


def eh_page_rpc_action(r):
    """
        Handle some player action (play, next, pause...)
    """
    r.send_response(200, 'OK')
    r.send_header('Pragma', 'no-cache')
    r.send_header('Content-type', 'text/plain')
    r.end_headers()
    r.wfile.write('OK')

    paths = r.path.split('?')
    path = paths[0]
    if path == '/rpc/action/play':
        if len(paths) > 1:
            args = cgi.parse_qs(paths[1])
            if args.has_key('f') and args['f'] != '' and APP.tracks is not None:
                songs = APP.tracks.get_songs()
                for track in songs:
                    if track.get_filename() == args['f'][0]:
                        print 'PLAYING filename=%s' % args['f'][0]
                        gobject.idle_add(APP.player.play_track, track, False, False)
                        return;
        if APP.player.is_paused():
            gobject.idle_add(APP.player.toggle_pause)
        else:
            gobject.idle_add(APP.player.play)
    elif path == '/rpc/action/next':
        gobject.idle_add(APP.player.next)
    elif path == '/rpc/action/previous':
        gobject.idle_add(APP.player.previous)
    elif path == '/rpc/action/pause':
        gobject.idle_add(APP.player.pause)
    elif path == '/rpc/action/stop':
        gobject.idle_add(APP.player.stop)
    elif path == '/rpc/action/seek':
        args = cgi.parse_qs(paths[1])
        if args.has_key('s') and args['s'] != '':
            gobject.idle_add(APP.player.seek, int(args['s'][0]), False)


def eh_page_playlist_list(r):
    """
        Return track informations from current playlist
    """
    r.send_response(200, 'OK')
    r.send_header('Pragma', 'no-cache')
    r.send_header('Content-type', 'text/xml; charset=utf-8')
    r.end_headers()

    r.wfile.write('<?xml version="1.0"?>')
    r.wfile.write('<playlist>')
    if APP.tracks is not None:
        songs = APP.tracks.get_songs()
        for song in songs:
            ti = TrackInfo(song, playlistmode=True)
            r.wfile.write('<track>')
            for key in ti.infos:
                r.wfile.write(tag(key, ti.infos[key]))
            r.wfile.write('</track>')
    r.wfile.write('</playlist>')



# -----------------------------------------------------------------------------
# HTTP server
# -----------------------------------------------------------------------------

eh_pages = {
        # Dynamic content
        '/image/cover/current':	eh_page_cover_current,

        # RPC url
        '/rpc/current':			eh_page_rpc_current,
        '/rpc/action/play':		eh_page_rpc_action,
        '/rpc/action/stop':		eh_page_rpc_action,
        '/rpc/action/next':		eh_page_rpc_action,
        '/rpc/action/previous':	eh_page_rpc_action,
        '/rpc/action/pause':	eh_page_rpc_action,
        '/rpc/action/seek':		eh_page_rpc_action,
        '/rpc/playlist/list':	eh_page_playlist_list,

        # Data
        '/':					eh_page_file,
        '/index.html':			eh_page_file,
        '/exaile.css':			eh_page_file,
        '/exaile.js':			eh_page_file,
        '/prototype.js':		eh_page_file,
        '/btn-play.png':		eh_page_file,
        '/btn-stop.png':		eh_page_file,
        '/btn-pause.png':		eh_page_file,
        '/btn-previous.png':	eh_page_file,
        '/btn-next.png':		eh_page_file,
        '/btn-reload.png':		eh_page_file,
        '/star.png':			eh_page_file,
        '/loading.gif':			eh_page_file,
        '/bg-trans.png':		eh_page_file,
    }

class ExaileHttpRequestHandler(BaseHTTPRequestHandler):

    def do_GET(self):

        path = self.path.split('?')[0]
        if eh_pages.has_key(path):
            eh_pages[path](self)
        else:
            self.send_error(404,'Not Found')
            return

        return

    @staticmethod
    def serve_forever(port):
        HTTPServer(('', port), ExaileHttpRequestHandler).serve_forever()

class ExaileHttpThread(threading.Thread):

    def __init__(self):
        threading.Thread.__init__(self)

    def run(self):
        print 'HTTP> http://127.0.0.1:%d/' % DEFAULT_PORT
        ExaileHttpRequestHandler.serve_forever(DEFAULT_PORT)



# -----------------------------------------------------------------------------
# Exaile interface
# -----------------------------------------------------------------------------

def initialize():
    global eh_thread
    eh_thread = ExaileHttpThread()
    eh_thread.setDaemon(True)
    eh_thread.start()
    return True

def destroy():
    global eh_thread
    if eh_thread is not None:
        del eh_thread
        eh_thread = None



# Copyright (C) 2006 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 1, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import os, re, urllib
from gettext import gettext as _, ngettext
import operator
import pygtk
pygtk.require('2.0')
import gtk, pango
from xl import xlmisc, common, library, burn, media
import xl.path
import editor, information

# creates the rating images for the caller
def create_rating_images(caller):
    """
        Called to (re)create the pixmaps used for the Rating column.
    """
    if (caller.rating_width != caller.old_r_w and caller.rating_width != 0):
        caller.rating_images = []
        star_size = caller.rating_width / 5

        star = gtk.gdk.pixbuf_new_from_file_at_size(
            xl.path.get_data('images', 'star.png'), star_size, star_size)

        full_image = gtk.gdk.Pixbuf(gtk.gdk.COLORSPACE_RGB, True, 8, caller.rating_width, star_size)
        full_image.fill(0xffffff00) # transparent white
        for x in range(0, 5):
            star.copy_area(0, 0, star_size, star_size, full_image, star_size * x, 0)
        caller.rating_images.insert(0, full_image)
        for x in range(5, 0, -1):
            this_image = gtk.gdk.Pixbuf(gtk.gdk.COLORSPACE_RGB, True, 8, caller.rating_width, star_size)
            this_image.fill(0xffffff00) # transparent white
            full_image.copy_area(0, 0, int(x * star_size), star_size, this_image, 0, 0)
            caller.rating_images.insert(0, this_image)
        caller.old_r_w = caller.rating_width

class Column(object):
    __slots__ = ['id', 'display', 'size']
    def __init__(self, id, display, size):
        self.id = id
        self.display = display
        self.size = size
    def __repr__(self):
        return 'Column(%s, %s, %s)' % (`self.id`, `self.display`, `self.size`)

class TracksListCtrl(gtk.VBox):
    """
        Represents the track/playlist table
    """
    rating_images = []
    rating_width = 64   # some default value
    old_r_w = -1
    row_height = 16
    
    COLUMNS = [
        Column('track', _('#'), 30),
        Column('title', _('Title'), 200),
        Column('artist', _('Artist'), 150),
        Column('album', _('Album'), 150),
        Column('length', _('Length'), 50),
        Column('disc_id', _('Disc'), 30),
        Column('rating', _('Rating'), 64),
        Column('year', _('Year'), 50),
        Column('genre', _('Genre'), 100),
        Column('bitrate', _('Bitrate'), 30),
        Column('io_loc', _('Location'), 200),
        Column('filename', _('Filename'), 200),
        Column('playcount', _('Playcount'), 50),
    ]

    COLUMN_IDS = []
    column_by_id = {}
    column_by_display = {}
    for col in COLUMNS:
        COLUMN_IDS.append(col.id)
        column_by_id[col.id] = col
        column_by_display[col.display] = col

    default_column_ids = ['track', 'title', 'album', 'artist', 'length']

    prep = "track"
    type = 'track'

    def __init__(self, exaile, queue=False):
        """
            Expects an exaile instance, the parent window, and whether or not
            this is a queue window.  If it's not a queue window, the track
            popup menu will be initialized.
        """
        gtk.VBox.__init__(self)
        self.exaile = exaile
        self.list = xlmisc.DragTreeView(self)
        self.list.set_rules_hint(True)
        self.list.set_enable_search(True)
        self.songs = library.TrackData()

        self.scroll = gtk.ScrolledWindow()
        self.scroll.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        self.scroll.add(self.list)
        self.pack_start(self.scroll, True, True)
        self.scroll.show_all()
        selection = self.list.get_selection()
        selection.set_mode(gtk.SELECTION_MULTIPLE)
        self.playimg = self.exaile.window.render_icon('gtk-media-play',
            gtk.ICON_SIZE_SMALL_TOOLBAR)
        self.playimg = self.playimg.scale_simple(18, 18,
            gtk.gdk.INTERP_BILINEAR)
        self.pauseimg = self.exaile.window.render_icon('gtk-media-pause',
            gtk.ICON_SIZE_SMALL_TOOLBAR)
        self.pauseimg = self.pauseimg.scale_simple(18, 18,
            gtk.gdk.INTERP_BILINEAR)

        self.db = exaile.db
        self.inited = False
        self.queue = queue
        self.playlist = None
        self.tpm = None
        self.plugins_item = None
        self.setup_columns()
        self.setup_events()
        
        create_rating_images(self)

        self.show()

    def close_page(self):
        """
            Called when this page in the notebook is closed
        """
        #Save the playlist order to the db
        cur = self.db.cursor()
        sort_index = 0;
        for track in self.songs:
            if track.type == 'stream': continue
            path_id = library.get_column_id(self.db, 'paths', 'name', track.loc)
            self.db.execute("UPDATE playlist_items SET sort_index=? " \
            "WHERE playlist_items.path=?" 
                            , (sort_index,path_id) )
            sort_index = sort_index + 1
        self.db.commit()

    def drag_data_received(self, tv, context, x, y, selection, info, etime):
        """
            Called when data is recieved
        """
        self.list.unset_rows_drag_dest()
        self.list.drag_dest_set(gtk.DEST_DEFAULT_ALL, 
            self.list.targets,
            gtk.gdk.ACTION_COPY|gtk.gdk.ACTION_MOVE)

        loc = list(selection.get_uris())
        counter = 0
 
        if context.action != gtk.gdk.ACTION_MOVE:
            self.exaile.status.set_first(
            _("Scanning and adding tracks to current playlist..."))
        xlmisc.finish()

        # first, check to see if they dropped a folder
        copy = loc[:]
        for l in copy:
            l = urllib.unquote(l)
            if os.path.isdir(l.replace("file://", "")):
                # in this case, it is a folder
                for root, dirs, files in os.walk(l.replace("file://", '')):
                    files.sort()
                    for file in files:
                        (stuff, ext) = os.path.splitext(file)
                        if ext.lower() in media.SUPPORTED_MEDIA:
                            loc.append(urllib.quote(os.path.join(root, file)))

        drop_info = tv.get_dest_row_at_pos(x, y)
        if drop_info:
            path, position = drop_info
            iter = self.model.get_iter(path)
            if (position == gtk.TREE_VIEW_DROP_BEFORE or
                position == gtk.TREE_VIEW_DROP_INTO_OR_BEFORE):
                first = False
            else:
                first = True

        for l in loc:
            l = l.replace("file://", "")
            l = urllib.unquote(l)
            m = re.search(r'^device_(\w+)://', l)
            if m:
                song = self.exaile.device_panel.get_song(l)
            else:
                song = library.read_track(self.exaile.db, self.exaile.all_songs, l)
                if not song:
                    # check plugins
                    song = self.get_plugin_track(l)

            if not song: continue
            if song in self.songs:
                # If the song is in drag_delete, we should remove it now
                # so it gets added properly in its new position.  Otherwise,
                # we are adding a track to the playlist that is already there,
                # and should do nothing.
                if song in self.drag_delete:
                    index = self.songs.index(song)
                    itera = self.model.get_iter((index,))
                    self.model.remove(itera)
                    self.songs.remove(song)
                    self.drag_delete.remove(song)
                    self.update_songs()
                else:
                    continue


            if not drop_info:
                # if there's no drop info, just add it to the end
                if song: self.append_song(song)
            else:
                if not first:
                    first = True
                    ar = self.get_ar(song)
                    if self.model.iter_is_valid(iter):
                        iter = self.model.insert_before(iter, ar)
                    else:
                        iter = self.model.append(ar)
                else:
                    ar = self.get_ar(song)
                    path = self.model.get_path(iter)
                    if self.model.iter_is_valid(iter):
                        iter = self.model.insert_after(iter, ar)
                    else:
                        iter = self.model.append(ar)

                if counter >= 20:
                    xlmisc.finish()
                    counter = 0
                else:
                    counter += 1
                    
                # Update here in case multiple tracks are being dragged.
                self.update_songs()
                
            # For some reason songs not in the database will always
            # show up as not in playlist_songs even if it is already there.
            # So instead of just appending if song is in self.playlist_songs, 
            # we compare track.loc to song.loc.  If that's there, we assign
            # song to the already existing track, otherwise, we append the
            # new song like normal.
            found_track = False
            for track in self.playlist_songs:                
                if track.loc == song.loc:
                    song = track
                    found_track = True
            if not found_track:
                self.playlist_songs.append(song)

        # Now that we've inserted new rows for the dragged tracks
        # we need to delete the original rows and track entries.
        if context.action == gtk.gdk.ACTION_MOVE:
            for track in self.drag_delete:
                index = self.songs.index(track)
                iter = self.model.get_iter((index,))
                self.model.remove(iter)
                self.songs.remove(track)
            self.drag_delete = []

        if context.action == gtk.gdk.ACTION_MOVE:
            context.finish(True, True, etime)
        else:
            context.finish(True, False, etime)
        self.update_songs()
        if self.type != 'queue':
            self.exaile.update_songs(self.songs, False)
        self.exaile.status.set_first(None)

    def get_plugin_track(self, loc):
        """
            Checks to see if a track is in the available plugin list
        """
        for k, v in self.exaile.plugin_tracks.iteritems():
            tr = v.for_path(loc)
            if tr: return tr

        return None

    def update_songs(self):
        """
            Updates the songs for the new order
        """
        songs = library.TrackData()
        iter = self.model.get_iter_first()
        if not iter: return
        while True:
            song = self.model.get_value(iter, 0)
            songs.append(song)
            iter = self.model.iter_next(iter)
            if not iter: break

        self.songs = songs

    def ensure_visible(self, track):
        """
            Scrolls to a track
        """
        if not track in self.songs: return
        index = self.songs.index(track)
        path = (index,)
        self.list.scroll_to_cell(path)
        self.list.set_cursor(path)
    
    def get_songs(self):
        """
            Returns the tracks displayed in this list
        """
        return self.songs

    def get_next_track(self, song):
        """
            Gets the next track
        """
        if song is None:
            index = -1
        else: 
            try:
                index = self.songs.index(song)
            except ValueError:
                return None
        path = (index + 1,)
        try:
            iter = self.model.get_iter(path)
        except ValueError:
            return None
        if not iter: return None
        return self.model.get_value(iter, 0)

    def get_previous_track(self, song):
        """
            Gets the previous track
        """
        if not song in self.songs: return None
        index = self.songs.index(song)
        path = (index - 1,)
        try:
            iter = self.model.get_iter(path)
        except ValueError:
            return None
        if not iter: return None
        return self.model.get_value(iter, 0)

    def get_selected_track(self):
        """
            Returns the selected track
        """
        selection = self.list.get_selection()
        (model, paths) = selection.get_selected_rows()
        if not paths: return
        iter = model.get_iter(paths[0])
        return model.get_value(iter, 0)

    def get_selected_tracks(self):
        """
            Gets the selected tracks in the tree view
        """
        selection = self.list.get_selection()
        (model, paths) = selection.get_selected_rows()
        songs = library.TrackData()
        for path in paths:
            iter = self.model.get_iter(path)
            song = self.model.get_value(iter, 0)
            songs.append(song)

        return songs

    def setup_events(self):
        """
            Sets up various events
        """
        if self.type != 'queue':
            self.list.connect('row-activated', lambda *e: self.exaile.player.play())
        self.list.connect('key_release_event', self.key_released)

    def drag_get_data(self, treeview, context, selection, target_id, etime):
        """
            Called when a drag source wants data for this drag operation
        """

        loc = []
        delete = []
        sel = self.list.get_selection()
        (model, paths) = sel.get_selected_rows()
        for path in paths:
            iter = self.model.get_iter(path)
            song = self.model.get_value(iter, 0) 

            if song.type == 'device':
                device_name = self.exaile.device_panel.get_driver_name()
                loc.append("device_%s://%s" % (device_name, urllib.quote(str(song.loc))))
            else:
                loc.append(urllib.quote(str(song.loc)))
            delete.append(song)
            
        # We can't remove the tracks in delete until the drag operation is
        # complete, because doing so causes strange dragging behavior.
        # So we just make the list available to the drag_data_recieved 
        # function to make use of.
        self.drag_delete = delete
        selection.set_uris(loc)

    def get_iter(self, song):
        """
            Returns the path for a song
        """
        iter = self.model.get_iter_first()
        while True:
            s = self.model.get_value(iter, 0)
            if s == song: return iter

            iter = self.model.iter_next(iter)
            if not iter: break

        return None

    def key_released(self, widget, event):
        """
            Called when someone presses a key
        """
        
        selection = self.list.get_selection()
        (model, pathlist) = selection.get_selected_rows()

        # Delete
        if event.keyval == gtk.keysyms.Delete:
            delete = []
            for path in pathlist:
                iter = model.get_iter(path)
                track = model.get_value(iter, 0)
                delete.append(track)

            for track in delete:
                iter = self.get_iter(track)
                model.remove(iter)
                self.songs.remove(track)
                if track in self.playlist_songs:
                    self.playlist_songs.remove(track)

            if pathlist and self.songs:
                path = pathlist[0]
                if path[0] >= len(self.songs): path = (path[0] - 1,)
                selection.select_path(path)

            update_queued(self.exaile)
            self.queue_draw()

    def setup_model(self, map):
        """
            Gets the array to build the two models
        """
        ar = [object, gtk.gdk.Pixbuf, gtk.gdk.Pixbuf]

        for item in map:
            if item == "track":
                ar.append(int)
            else:
                ar.append(str)

        self.model = gtk.ListStore(*ar)
        self.model_blank = gtk.ListStore(*ar)
        self.list.set_model(self.model)
        self.set_songs(self.songs)

    def setup_columns(self):
        """
            Sets up the columns for this table
        """

        self._col_count = 0
        self._length_id = -1

        col_ids = self.exaile.settings.get_list("ui/col_order", None)
        search_column = self.exaile.settings.get_str("ui/search_column","Title")
        
        cols = None
        if col_ids:
            cols = []
            for col_id in col_ids[:]:
                col = self.column_by_id.get(col_id)
                if col: # Good entries only
                    cols.append(col)
                else:
                    col_ids.remove(col_id)
        if cols:
            for col in self.COLUMNS:
                if not col in cols:
                    cols.append(col)
                    col_ids.append(col.id)
        else:
            cols = self.COLUMNS[:]
            col_ids = self.COLUMN_IDS[:]

        self.append_map = col_ids
        self.setup_model(col_ids)

        count = 3
        first_col = True
        columns_settings = self.exaile.settings.get_list("ui/%s_columns" % self.prep)

        for col_struct in cols:
            # get cell renderer
            cellr = gtk.CellRendererText()
            if col_struct.id == 'rating':
                cellr = gtk.CellRendererPixbuf()
                cellr.set_property("follow-state", False)

            if col_struct.id in columns_settings:
                if first_col:
                    first_col = False
                    pb = gtk.CellRendererPixbuf()
                    pb.set_fixed_size(20, 20)
                    pb.set_property('xalign', 0.0)
                    stop_pb = gtk.CellRendererPixbuf()
                    stop_pb.set_fixed_size(12, 12)
                    col = gtk.TreeViewColumn(col_struct.display)
                    col.pack_start(pb, False)
                    col.pack_start(stop_pb, False)
                    col.pack_start(cellr, True)
                    col.set_attributes(cellr, text=count)
                    col.set_attributes(pb, pixbuf=1)
                    col.set_attributes(stop_pb, pixbuf=2)
                    col.set_cell_data_func(pb, self.icon_data_func)
                    col.set_cell_data_func(stop_pb, self.stop_icon_data_func)
                else:
                    col = gtk.TreeViewColumn(col_struct.display, cellr, text=count)

                if col_struct.id == 'length':
                    col.set_cell_data_func(cellr, self.length_data_func)
                elif col_struct.id == 'track':
                    col.set_cell_data_func(cellr, self.track_data_func)
                elif col_struct.id == 'disc_id':
                    col.set_cell_data_func(cellr, self.disc_data_func)
                elif col_struct.id == 'rating':
                    col.set_attributes(cellr, pixbuf=1)
                    col.set_cell_data_func(cellr, self.rating_data_func)
                else:
                    col.set_cell_data_func(cellr, self.default_data_func)

                setting_name = "ui/%scol_width_%s" % (self.prep, col_struct.id)
                width = self.exaile.settings.get_int(setting_name, 
                    col_struct.size)
                col.set_fixed_width(width)

                resizable = self.exaile.settings.get_boolean('ui/resizable_cols',
                    False)

                if self.type != 'queue':
                    col.connect('clicked', self.set_sort_by)
                    col.connect('notify::width', self.set_column_width)
                    col.set_clickable(True)
                    col.set_reorderable(True)
                    col.set_resizable(False)
                    col.set_sort_indicator(False)

                if not resizable:
                    if col_struct.id in ('title', 'artist', 'album', 'io_loc', 'genre'):
                        if col_struct.id != 'genre': 
                            col.set_expand(True)
                            col.set_fixed_width(1)
                        else:
                            col.set_fixed_width(80)
                        col.set_sizing(gtk.TREE_VIEW_COLUMN_FIXED)
                        cellr.set_property('ellipsize', pango.ELLIPSIZE_END)
                    else:
                        col.set_sizing(gtk.TREE_VIEW_COLUMN_AUTOSIZE)
                else:
                    col.set_resizable(True)
                    col.set_sizing(gtk.TREE_VIEW_COLUMN_FIXED)

                if col_struct.id in ('track', 'playcount'):
                    cellr.set_property('xalign', 1.0)
                    
                # Update which column to search for when columns are changed
                if col_struct.display == search_column:
                    self.list.set_search_column(count)
                col.set_widget(gtk.Label(col_struct.display))
                col.get_widget().show()
                self.list.append_column(col)
                col.get_widget().get_ancestor(gtk.Button).connect('button_press_event', self.press_header)
            count = count + 1
        self.changed_id = self.list.connect('columns-changed', self.column_changed)
    
    def press_header(self, widget, event):
        if event.button != 3:
            return False
        menu = self.exaile.xml.get_widget('columns_menu_menu')
        menu.popup(None, None, None, event.button, event.time)
        return True

    def header_toggle(self, menuitem, column):
        column.set_visible(not column.get_visible())

    def column_changed(self, *e):
        """
            Called when columns are reordered
        """
        self.list.disconnect(self.changed_id)
        cols = []
        for col in self.list.get_columns():
            cols.append(self.column_by_display[col.get_title()].id)
            self.list.remove_column(col)

        self.exaile.settings['ui/col_order'] = cols
        self.setup_columns()

    def set_column_width(self, col, stuff=None):
        """
            Called when the user resizes a column
        """
        col_struct = self.column_by_display[col.get_title()]
        name = "ui/%scol_width_%s" % (self.prep, col_struct.id)
        self.exaile.settings[name] = col.get_width()
        if col_struct.id == 'rating':
            self.rating_width = min(col.get_width(), self.row_height * 4)
            create_rating_images(self)

    def icon_data_func(self, col, cell, model, iter):
        """
            Sets track status (playing/paused/queued) icon
        """

        item = model.get_value(iter, 0)
        image = None

        if item == self.exaile.player.current:
            if self.exaile.player.is_playing():
                image = self.playimg
            elif self.exaile.player.is_paused():
                image = self.pauseimg
        elif item in self.exaile.player.queued:
            index = self.exaile.player.queued.index(item)
            image = xlmisc.get_text_icon(self.exaile.window,
                str(index + 1), 18, 18)

        cell.set_property('pixbuf', image)

    def stop_icon_data_func(self, col, cell, model, iter):
        """
            Sets "stop after this" icon
        """

        item = model.get_value(iter, 0)
        image = None
        
        if item == self.exaile.player.stop_track:
            image = self.exaile.window.render_icon('gtk-stop', 
                gtk.ICON_SIZE_MENU) 
            image = image.scale_simple(12, 12, gtk.gdk.INTERP_BILINEAR)
        
        cell.set_property('pixbuf', image)  

    def rating_data_func(self, col, cell, model, iter):
        item = model.get_value(iter, 0)
        if not item._rating: return
        try:
            idx = item._rating - 1
            cell.set_property('pixbuf', self.rating_images[idx])
        except IndexError:
            if idx > 5: idx = 5
            elif idx < 0: idx = 0

    def set_cell_weight(self, cell, item):
        """
            Sets a CellRendererText's "weight" property according to whether
            `item` is the currently playing track.
        """
        if item == self.exaile.player.current:
            weight = pango.WEIGHT_HEAVY
        else:
            weight = pango.WEIGHT_NORMAL
        cell.set_property('weight', weight)

    def disc_data_func(self, col, cell, model, iter):
        """
            Formats the disc
        """
        item = model.get_value(iter, 0)
        if item.disc_id is None or item.disc_id == -1 or \
            item.type == 'podcast':
            cell.set_property('text', '')
        else:
            cell.set_property('text', item.disc_id)
        self.set_cell_weight(cell, item)

    def filename_data_func(self, col, cell, model, iter):
        """
            Shows the filename of the track
        """
        item = model.get_value(iter, 0)
        cell.set_property('text', os.path.basename(item.io_loc))
        self.set_cell_weight(cell, item)

    def length_data_func(self, col, cell, model, iter):
        """ 
            Formats the track length
        """
        item = model.get_value(iter, 0)
        if item.type == 'podcast':
            text = item.length
            if item.length == "0:00":
                text = "N/A"
        elif item.type == 'stream':
            text = ''
        else:
            seconds = item.duration
            text = "%s:%02d" % (seconds / 60, seconds % 60)
        cell.set_property('text', text)
        self.set_cell_weight(cell, item)

    def track_data_func(self, col, cell, model, iter):
        """
            Track number
        """
        item = model.get_value(iter, 0)
        if item.track is None or item.track == -1 or \
            item.type == 'podcast':
            cell.set_property('text', '')
        else:
            cell.set_property('text', item.track)
        self.set_cell_weight(cell, item)

    def default_data_func(self, col, cell, model, iter):
        """
            For use in CellRendererTexts that don't have special data funcs.
        """
        item = model.get_value(iter, 0)
        self.set_cell_weight(cell, item)

    # sort functions courtesy of listen (http://listengnome.free.fr), which
    # are in turn, courtesy of quodlibet.  
    def set_sort_by(self, column):
        """
            Sets the sort column
        """
        title = column.get_title()
        count = 0
        for col in self.list.get_columns():
            if title == col.get_title():
                order = column.get_sort_order()
                if (not column.get_sort_indicator() or 
                    order == gtk.SORT_DESCENDING):
                    order = gtk.SORT_ASCENDING
                else:
                    order = gtk.SORT_DESCENDING
                col.set_sort_indicator(True)
                col.set_sort_order(order)
            else:
                col.set_sort_indicator(False)

        self.songs = self.reorder_songs(self.songs)
        self.set_songs(self.songs)

    def reorder_songs(self, songs):
        """
            Resorts all songs
        """
        attr, reverse = self.get_sort_by()

        def the_strip(tag):
            return spec_strip(library.the_cutter(tag))
        def spec_strip(tag):
            return library.lstrip_special(tag)

        if attr in ('album', 'title'):
            get_key = lambda track: spec_strip(getattr(track, attr).lower())
        elif attr == 'artist':
            get_key = lambda track: the_strip(track.artist.lower())
        elif attr == 'length':
            get_key = lambda track: track.get_duration()
        else:
            get_key = lambda track: getattr(track, attr)

        s = [
            (get_key(track),
            track)
            for track in songs]

        s  = sorted(s, key=operator.itemgetter(0), reverse=reverse)

        songs = [track[-1] for track in s]
        return songs

    def get_sort_by(self):
        """
            Gets the sort order
        """
        for col in self.list.get_columns():
            if col.get_sort_indicator():
                return (self.column_by_display[col.get_title()].id,
                    col.get_sort_order() == gtk.SORT_DESCENDING)
        return 'album', False

    def update_col_settings(self):
        """
            Updates the settings for a specific column
        """
        self.list.disconnect(self.changed_id)
        columns = self.list.get_columns()
        for col in columns:
            self.list.remove_column(col)

#        self.tree_lost_focus(None, None)
        self.setup_columns()
        self.list.queue_draw()

    def set_songs(self, songs, update_playlist=True):
        """
            Sets the songs in this table (expects a list of tracks)
        """
        self.songs = library.TrackData()
        search_column = self.exaile.settings.get_str("ui/search_column","Title")
        # save sort indicators, because they get reset when you set the model
        indicators = {}
        self.model.clear()
        for col in self.list.get_columns():
            indicators[col.get_title()] = col.get_sort_indicator()

        self.list.set_model(self.model_blank)
        if update_playlist: self.playlist_songs = songs
        for song in songs:
            self.append_song(song,False)
        self.songs = songs;

        self.list.set_model(self.model)
        count = 3
        for col in self.list.get_columns():
            col.set_sort_indicator(indicators[col.get_title()])
            # We have to set search_column here because the model was cleared
            if col.get_title() == search_column:
                self.list.set_search_column(count)
            count = count + 1  

    def get_ar(self, song):
        """
            Creates the array to be added to the model in the correct order
        """
        ar = [song, None, None]
        for field in self.append_map:
            ar.append(getattr(song, field))
        return ar

    def append_song(self, song, append_to_list = True):
        """
            Adds a song to this view
        """
        ar = self.get_ar(song)

        self.model.append(ar)
        if append_to_list: self.songs.append(song)

    def update_iter(self, iter, song):
        """
            Updates the track at "iter"
        """
        ar = self.get_ar(song)
        self.model.insert_after(iter, ar)
        self.model.remove(iter)
    
    def update_song_list(self, song, orig_song):
        """
            Updates the in the song list with the new Track object
        """
        index = self.songs.index(orig_song)
        self.songs[index] = song

    def refresh_row(self, song):
        """
            Refreshes the text for the specified row
        """
        selection = self.list.get_selection()
        model, paths = selection.get_selected_rows()
        iter = self.model.get_iter_first()
        if not iter: return
        while True:
            check = self.model.get_value(iter, 0)
            if not check: break
            if check == song or check.io_loc == song.io_loc:
                self.update_iter(iter, song)
                self.update_song_list(song, check)
                break
            iter = self.model.iter_next(iter)
            if not iter: break
      
        if not paths: return
        for path in paths:
            selection.select_path(path)

    def setup_tracks_menu(self):
        """
            Sets up the popup menu for the tracks list
        """
        tpm = xlmisc.Menu()

        # if the menu already exists, remove the plugins menu from it
        if self.plugins_item:
            self.plugins_item.remove_submenu()
            
        self.tpm = tpm        
                
        songs = self.get_selected_tracks()
        t = songs[0].type

        n_selected = len(songs)

        pixbuf = xlmisc.get_text_icon(self.exaile.window, u'\u2610', 16, 16)
        icon_set = gtk.IconSet(pixbuf)
        
        factory = gtk.IconFactory()
        factory.add_default()        
        factory.add('exaile-queue-icon', icon_set)
        
        self.queue = tpm.append(_("Toggle Queue"), self.exaile.on_queue,
            'exaile-queue-icon')

        if n_selected == 1:
            self.stop_track = tpm.append(_("Toggle: Stop after this Track"), 
                self.exaile.on_stop_track, 'gtk-stop')
        tpm.append_separator()

        if t == 'podcast':
            tpm.append(_('Download Podcasts'), self.download_podcast,
            'gtk-save')
        tpm.append(_("Edit Track Information"), lambda e, f:
            editor.TrackEditor(self.exaile, self), 'gtk-edit')

        rm = xlmisc.Menu()

        for i in range(0, 5):
            if i == 0:
                item = rm.append('-', lambda w, e, i=i:
                    editor.update_rating(self, i))
            else:
                item = rm.append_image(self.rating_images[i],
                    lambda w, e, i=i: editor.update_rating(self, i))

        star_icon = gtk.gdk.pixbuf_new_from_file_at_size(
            xl.path.get_data('images', 'star.png'), 16, 16)
        icon_set = gtk.IconSet(star_icon)
        factory = gtk.IconFactory()
        factory.add_default()        
        factory.add('exaile-star-icon', icon_set)

        tpm.append_menu(_("Set Track Rating"), rm, 'exaile-star-icon')

        if t == 'file':
            bm = xlmisc.Menu()
            bm.append(ngettext("Burn Selected Track", "Burn Selected Tracks",
                n_selected), self.burn_selected, 'gtk-cdrom')
            bm.append(_("Burn Playlist"), self.burn_playlist, 'gtk-cdrom')
            tpm.append_menu(_('Burn'), bm, 'gtk-cdrom')
            if not burn.check_burn_progs():
                bm.set_sensitive(False)

        if t == 'cd':
            im = xlmisc.Menu()
            im.append(ngettext("Import Selected Track",
                "Import Selected Tracks", n_selected),
                self.import_selected, 'gtk-cdrom')
            im.append(_('Import CD'), self.import_cd, 'gtk-cdrom')
            tpm.append_menu(_('Import'), im, 'gtk-cdrom')

        if t != 'cd':
            if not songs or not t == 'stream':
                pm = xlmisc.Menu()
                self.new_playlist = pm.append(_("New Playlist"),
                    self.exaile.playlists_panel.on_add_playlist, 'gtk-new')
                pm.append_separator()
                rows = self.db.select("SELECT name FROM playlists WHERE type=0 ORDER BY"
                    " name")
                for row in rows:
                    pm.append(row[0], self.exaile.playlists_panel.add_items_to_playlist)

                tpm.append_menu(_("Add to Playlist"), pm, 'gtk-add')
            else:
                self.playlists_menu = xlmisc.Menu()
                all = self.db.select("SELECT name FROM radio "
                    "ORDER BY name")
                for row in all:
                    i = self.playlists_menu.append(row[0],
                        self.exaile.pradio_panel.add_items_to_station)

                self.playlists_menu.append_separator()
                self.new_playlist = self.playlists_menu.append(_("New Station"),
                    self.exaile.pradio_panel.on_add_to_station, 'gtk-new')

                tpm.append_menu(_("Add to Saved Stations"),
                    self.playlists_menu, 'gtk-add')


        if n_selected == 1:
            info = tpm.append(_("Information"), self.get_track_information,
                'gtk-info')
        tpm.append_separator()

        if n_selected == 1 and self.get_selected_track() \
            and self.get_selected_track().type == 'lastfm':
            lfm = xlmisc.Menu()
            lfm.append('Ban', lambda *e: self.send_lastfm_command('ban'))
            lfm.append('Love', lambda *e: self.send_lastfm_command('love'))
            lfm.append('Skip', lambda *e: self.send_lastfm_command('skip'))
            tpm.append_menu(_("Last.FM Options"), lfm)
            tpm.append_separator()

        factory.add('exaile-track-icon', gtk.IconSet(
            gtk.gdk.pixbuf_new_from_file(xl.path.get_data('images',
            'track.png'))))
        # TRANSLATORS: Shows the selected track in the collection tree
        if n_selected == 1:
            tpm.append(_("Show in Collection"), self.show_in_collection,
                'exaile-track-icon')

        rm = xlmisc.Menu()
        rm.append(_("Remove Selected from Playlist"),
            self.delete_tracks, 'gtk-delete')
        rm.append(_("Blacklist Selected"),
            self.delete_tracks, 'gtk-delete', 'blacklist')
        rm.append(_("Delete Selected"),
            self.delete_tracks, 'gtk-delete', 'delete')

        tpm.append_menu(_("Remove"), rm, 'gtk-delete')

        # plugins menu items
        if self.exaile.plugins_menu.get_children():
            tpm.append_separator()

            # TRANSLATORS: Plugin submenu for the playlist
            self.plugins_item = tpm.append(_("Plugins"), None, 'gtk-execute')
            self.plugins_item.set_submenu(self.exaile.plugins_menu)

    def burn_selected(self, widget, event):
        burn.launch_burner(self.exaile.settings.get_str('burn_prog', burn.check_burn_progs()[0]), \
            self.exaile.tracks.get_selected_tracks())

    def show_in_collection(self, item, event):
        """
            Go to collection tree item corresponding to current track
        """
        track = self.get_selected_track()
        self.exaile.collection_panel.show_in_collection(track)

    def burn_selected(self, widget, event):
        burn.launch_burner(self.exaile.settings.get_str('burn_prog', burn.check_burn_progs()[0]), \
            self.exaile.tracks.get_selected_tracks())

    def burn_playlist(self, widget, event):
        burn.launch_burner(self.exaile.settings.get_str('burn_prog', burn.check_burn_progs()[0]), \
            self.songs)

    def import_selected(self, widget, event):
        self.exaile.importer.do_import(self.get_selected_tracks())

    def import_cd(self, widget, event):
        self.exaile.importer.do_import(self.songs)

    def send_lastfm_command(self, command):
        if self.exaile.player.lastfmsrc:
            self.exaile.player.lastfmsrc.control(command)

    def get_track_information(self, widget, event=None):
        """
            Shows the track information tab
        """
        t = self.get_selected_track()
        information.show_information(self.exaile, t)

    def button_press(self, button, event):
        """
            The popup menu that is displayed when you right click in the
            playlist
        """
        if not event.button == 3: 
            try: path, col, cellx, celly = self.list.get_path_at_pos(int(event.x),
            int(event.y))
            except: 
                xlmisc.log_exception()
                return True

            if col.get_title() == _('Rating'):
                count = int(float(cellx + 5) / (float(self.rating_width / 5)))
                print count
            return True
            print path, col, cellx, celly, col.get_title()
            
        selection = self.list.get_selection()
        self.setup_tracks_menu()

        self.tpm.popup(None, None, None, 0, gtk.get_current_event_time())
        if selection.count_selected_rows() <= 1: return False
        else: return True

    def load(self, genre, plugin):
        """
            Loads the track information from a cache file if it exists.
            If loading it fails, False is returned, else True is returned
        """
        cache = xl.path.get_config('cache')
        if not os.path.isdir(cache): os.mkdir(cache)
        f = "%s%sradioplugin_%s_%s.tab" % (cache, os.sep, plugin, genre)
        if not os.path.isfile(f): return False
        songs = []
        try:
            f = open(f, "r")
            for line in f.readlines():
                line = line.strip()
                fields = line.split("\t")
                info = ({
                    "album": fields[0],
                    "artist": fields[1],
                    "title": fields[2],
                    "loc": fields[3],
                    "bitrate": fields[4].replace("k", "")
                })

                track = media.Track()
                track.set_info(**info)
                track.type = 'stream'
                songs.append(track)
        except:
            xlmisc.log_exception()
            return False
        self.set_songs(songs)
        self.playlist_songs = songs
        return True

    def save(self, genre, plugin):
        """
            Saves the track information to a track file.
        """
        cache = xl.path.get_config('cache')
        if not os.path.isdir(cache): os.mkdir(cache)

        try:
            f = open("%s%sradioplugin_%s_%s.tab" % (cache, os.sep, plugin, genre), "w")
            for track in self.songs:
                f.write("%s\t%s\t%s\t%s\t%s\n" % (track.album, track.artist, track.title,
                    track.loc, track.bitrate))
            f.close()
        except IOError:
            xlmisc.log('Could not save station list')

    def download_podcast(self, event, type):
        """
            Downloads the selected podcasts
        """
        songs = self.get_selected_tracks()
        for song in songs:
            if song.type == 'podcast':
                self.exaile.pradio_panel.add_podcast_to_queue(song)

    def delete_tracks(self, event, type): 
        """
            Deletes tracks, or just removes them from the playlist
        """
        deleting = False
        blacklisting = False
        if type == 'delete': deleting = True
        if type == 'blacklist': blacklisting = True
        delete = []
        delete_confirmed = False

        device_delete = []
        if deleting:
            result = common.yes_no_dialog(self.exaile.window, _("Are you sure "
            "you want to permanently remove the selected tracks from disk?"))
            if result == gtk.RESPONSE_YES: delete_confirmed = True
            else: return
        cur = self.db.cursor()

        error = ""

        if not deleting or delete_confirmed:
            tracks = self.get_selected_tracks()
            
            # stores song before the first selected one so that it can be focused later
            if (self.songs[0] != tracks[0]):
                saved_song_iter = self.get_iter(self.get_previous_track(tracks[0]))
                saved_song_tree_path = self.model.get_path(saved_song_iter)
                scroll = True
            else:
                scroll = False

            
            for track in tracks:
                delete.append(track)

            while len(delete) > 0:
                track = delete.pop()
                path_id = library.get_column_id(self.db, 'paths', 'name', track.loc)
                self.exaile.playlist_songs.remove(track)
                if track == self.exaile.player.stop_track:
                    self.exaile.player.stop_track = None
                try: self.songs.remove(track)
                except ValueError: pass

                # I use exceptions here because the "in" operator takes
                # time that I'm sure has to be repeated in the "remove" method
                # (or at least the "index" method is called, which probably
                # ends up looping until it finds it anyway
                try: self.exaile.songs.remove(track)
                except ValueError: pass

                if deleting or blacklisting:
                    try: self.exaile.all_songs.remove(track)
                    except ValueError: pass

                if deleting:
                    xlmisc.log("Deleting %s" % track.loc)

                    if track.type == 'device':
                        xlmisc.log("Device track detected")
                        device_delete.append(track)
                        continue
                    db = self.db
                    try:
                        if track.type == 'podcast':
                            if track.download_path:
                                os.remove(track.download_path)
                        else:
                            if os.path.isfile(track.loc): 
                                os.remove(track.loc)
                    except OSError:
                        common.error(self.exaile.window, "Could not delete '%s' - "\
                            "perhaps you do not have permissions to do so?"
                            % track.loc)
                    cur.execute("DELETE FROM tracks WHERE path=?", (path_id,))
                else: 
                    # execute if this is "remove from playlist" or "blacklist"
                    playlist = self.playlist
                    if playlist != None:
                        if track.type == 'stream':
                            t = "radio"; p = "url"
                        else:
                            t = "playlists"; p = "path"

                        playlist_id = library.get_column_id(self.db, t, 'name',
                            playlist)

                        if t == 'playlists':
                            t = 'playlist'
                        cur.execute("DELETE FROM %s_items WHERE "
                            "path=? AND %s=?" % (t, t),
                            (path_id, playlist_id))
                    if blacklisting:
                        cur.execute("UPDATE tracks SET blacklisted=1 "
                            "WHERE path=?", (path_id,))

            cur.close()
            if error:
                common.scrolledMessageDialog(self.exaile.window,
                    error, _("The following errors did occur"))
            self.exaile.collection_panel.track_cache = dict()
            self.set_songs(self.songs)
            
            # move to old position
            if scroll:
                self.list.scroll_to_cell(saved_song_tree_path, use_align=True)

            if blacklisting: self.exaile.show_blacklist_manager(False)
            if device_delete:
                self.exaile.device_panel.remove_tracks(device_delete)

class BlacklistedTracksList(TracksListCtrl):
    """
        Shows a list of blacklisted tracks
    """
    blacklist = None
    type = 'blacklist'
    def __init__(self, exaile):
        """
            Initializes the tracks list
        """
        TracksListCtrl.__init__(self, exaile, False)

    def setup_tracks_menu(self):
        """
            Sets up the popup menu for the tracks list
        """
        TracksListCtrl.setup_tracks_menu(self)
        self.deblacklist = self.tpm.append(_("Remove from Blacklist"), 
            self.de_blacklist)

    def de_blacklist(self, item, event):
        """
            Removes items from the blacklist
        """
        songs = self.get_selected_tracks()
        cur = self.db.cursor()
        remove = []
        for track in songs:
            remove.append(track)

        for track in remove:
            self.playlist_songs.remove(track)
            try: self.exaile.songs.remove(track)
            except: pass
            path_id = library.get_column_id(self.db, 'paths', 'name', track.loc)
            cur.execute("UPDATE tracks SET blacklisted=0 WHERE path=?",
                (path_id,)) 
            if not track in self.exaile.all_songs:
                self.exaile.all_songs.append(track)

        self.set_songs(self.exaile.songs)
        self.exaile.collection_panel.track_cache = dict()

class QueueManager(TracksListCtrl):
    """
        Represents all tracks present in the queue
    """
    type = 'queue'
    def __init__(self, exaile):
        """
            Initializes the queue manager
        """
        TracksListCtrl.__init__(self, exaile, True)

        buttons = gtk.HBox()
        buttons.set_border_width(5)
        self.clear = gtk.Button(None)
        self.clear.connect('clicked', self.clear_queue)
        self.clear.set_image(self.image('gtk-clear'))
        buttons.pack_end(self.clear, False, False)
        self.delete = gtk.Button(None)
        self.delete.set_image(self.image('gtk-delete'))
        self.delete.connect('clicked', self.delete_tracks)
        buttons.pack_end(self.delete, False, False)
        self.up = gtk.Button(None)
        self.up.set_image(self.image('gtk-go-up'))
        self.up.connect('clicked', self.move_up)

        buttons.pack_end(self.up, False, False)
        self.down = gtk.Button(None)
        self.down.set_image(self.image('gtk-go-down'))
        self.down.connect('clicked', self.move_down)
        buttons.pack_end(self.down, False, False)
        
        self.pack_start(buttons, False, True)
        buttons.show_all()

        self.set_songs(self.exaile.player.queued)

    def button_press(self, button, event):
        """
            Overridden method of TracksListCtrl.  Does nothing at all
        """
        pass

    def move_down(self, widget):
        """
            Moves the track down in the queue
        """
        selection = self.list.get_selection()
        (model, paths) = selection.get_selected_rows()

        try:
            num = paths[0][0]
        except IndexError:
            pass

        iter = model.get_iter(paths[0])

        queue = self.exaile.player.queued
        item = queue.pop(num)
        num += 1
        if num > len(queue):
            num = 0

        queue.insert(num, item)
        self.set_songs(queue)
        selection.select_path((num,))

    def move_up(self, widget):
        """
            Moves the track up in the queue
        """
        selection = self.list.get_selection()
        (model, paths) = selection.get_selected_rows()
        try:
            num = paths[0][0]
        except IndexError:
            pass
        iter = model.get_iter(paths[0])

        queue = self.exaile.player.queued
        item = queue.pop(num)
        num -= 1
        if num < 0:
            num = len(queue)

        queue.insert(num, item)
        self.set_songs(queue)
        selection.select_path((num,))
        
    def delete_tracks(self, item):
        """
            Deletes tracks from the queue
        """
        tracks = self.get_selected_tracks()
        for track in tracks:
            self.exaile.player.queued.remove(track)
            if track == self.exaile.player.stop_track:
                self.exaile.player.stop_track = None

        self.set_songs(self.exaile.player.queued)
        update_queued(self.exaile)

    def clear_queue(self, item):
        """
            Clears the queue
        """
        while self.exaile.player.queued:
            self.exaile.player.queued.pop()

        self.set_songs(self.exaile.player.queued)
        update_queued(self.exaile)

    def image(self, i):
        """
            Returns a stock image
        """
        return gtk.image_new_from_stock(i, gtk.ICON_SIZE_SMALL_TOOLBAR)

    def set_songs(self, songs):
        """
            Sets the tracks for the queue manager, and then disables header
            sorting
        """
        TracksListCtrl.set_songs(self, songs)

    def drag_data_received(self, tv, context, x, y, selection, info, etime):
        """
            Proxy to recieved dragged items.  Calls the parent class, then
            updates the queue order based on the new order of the tracks
        """
        TracksListCtrl.drag_data_received(self, tv, context, x, y, selection, 
            info, etime)
        self.exaile.player.queued = []
        for song in self.songs:
            self.exaile.player.queued.append(song)

        update_queued(self.exaile)

def update_queued(exaile):
    """ 
        If the queue manager is showing, this updates it
    """
    nb = exaile.playlists_nb
    for i in range(0, nb.get_n_pages()):
        page = nb.get_nth_page(i)
        if isinstance(page, QueueManager):
            page.set_songs(exaile.player.queued)

    if exaile.player.queued:
        n = len(exaile.player.queued)
        exaile.queue_count_label.set_label(ngettext(": %d track queued",
            ": %d tracks queued", n) % n)
    else:
        exaile.queue_count_label.set_label("")




# Copyright (C) 2006 Adam Olsen 
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 1, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import os, zipfile, md5, random, time, re, sys
import xl.plugins as plugins
from gettext import gettext as _
random.seed(time.time())
from xl import common, xlmisc
from xl.panels import radio
import xl.path
import xl.media as media
import time, gtk, gobject, gtk.glade
import urllib

PLUGIN_NAME = _("LastFM Radio")
PLUGIN_AUTHORS = ['Adam Olsen <arolsen@gmail.com>']
PLUGIN_VERSION = "0.2.7"
PLUGIN_DESCRIPTION = _(r"""Allows for streaming via lastfm proxy.\n\nThis
plugin is very beta and still doesn't work perfectly.""")
PLUGIN_ENABLED = False
PLUGIN_ICON = None

TMP_DIR = None
PROXY = None
PLUGIN = None
HTTP_CLIENT = None
BUTTONS = []
TIPS = gtk.Tooltips()
GLADE_XML_STRING = None
CONS = plugins.SignalContainer()
SHOW_COVER = False
OLD_COVER_FUNC = None

def lastfm_error(message):
    common.error(APP.window, message)

class LastFMDriver(radio.RadioDriver):
    def __init__(self, panel):
        self.model = panel.model
        self.folder_icon = panel.folder
        self.note = panel.track
        self.tree = panel.tree
        self.panel = panel
        self.exaile = APP
        self.last_node = None
        self.no_new_page = True

        self.custom_urls = self.exaile.settings.get_list('custom_urls',
            plugin=plugins.name(__file__))

    def command(self, command):
        self.lfmcommand = command
        self.exaile.status.set_first(_("Running command: %s...") %
            command, 3500) 
        self.do_command()

    @common.threaded
    def do_command(self):
        HTTP_CLIENT.req(self.lfmcommand)

    def get_menu(self, item, menu):
        menu.append_separator()
        menu.append(_("Add Station"), self.on_add, 'gtk-add')
        if isinstance(item, radio.RadioGenre) and item.custom:
            menu.append(_("Remove Station"), self.on_remove, 'gtk-delete')
            self.last_menu_item = item
        menu.append_separator()
        menu.append(_("Play 'Similar Artist' Radio"), lambda *e:
            self.play_radio(
                _('Similar Artist'), 
                _('Artist'),
                "lastfm://artist/%s/similarartists"),
            'gtk-media-play')

        menu.append(_("Play 'Neighbour' Radio"), lambda *e:
            self.play_radio(
                _('Neighbour'), 
                _('User'),
                "lastfm://user/%s/personal"),
            'gtk-media-play')

        menu.append(_("Play 'Group' Radio"), lambda *e:
            self.play_radio(
                _('Group'),
                _("Group"),
                "lastfm://group/%s"),
            'gtk-media-play')

        menu.append(_('Play URL'), lambda *e: 
            self.play_radio(
                '', 
                _("URL"),
                'lastfm://%s'),
            'gtk-media-play')
            
        return menu

    def play_radio(self, title, field, url):
        """
            Opens a dialog and asks the user for a field, and plays the
            station
        """
        dialog = common.MultiTextEntryDialog(APP.window, title)
        dialog.add_field("%s: " % field)
        result = dialog.run()
        dialog.hide()

        if result == gtk.RESPONSE_OK:
            values = dialog.get_values()
            display = values[0]
            value = urllib.quote(values[0].replace("lastfm://", ''))

            item = radio.RadioGenre("%s: %s" % (field, display))
            item.lastfm_url = url % value
            self.load_genre(item)

    def on_add(self, *e):
        """
            Called when the user wants to add a station
        """
        dialog = common.MultiTextEntryDialog(APP.window, _("Add Station"))
        dialog.add_field(_("Name"))
        dialog.add_field(_("URL"))

        result = dialog.run()
        values = dialog.get_values()
        dialog.hide()
        if result == gtk.RESPONSE_OK:
            if values[1].find("://") == -1:
                values[1] = "lastfm://%s" % values[1]
            self.custom_urls.append(values)
            self.exaile.settings.set_list('custom_urls',
                self.custom_urls, plugin=plugins.name(__file__))

            item = radio.RadioGenre(values[0], self)
            item.lastfm_url = values[1]
            item.custom = True
            node = self.model.append(self.last_node, [self.note, item])
            item.node = node

    def on_remove(self, *e):
        item = self.last_menu_item
        node = item.node

        result = common.yes_no_dialog(APP.window, _("Are you sure you "
            "want to remove this item?"))
        if result == gtk.RESPONSE_YES:
            self.model.remove(node)
            
            new_custom = []
            for station, url in self.custom_urls:
                if item.name == station and item.lastfm_url == url:
                    continue
                else:
                    new_custom.append([station, url])

            self.custom_urls = new_custom
            self.exaile.settings.set_list('custom_urls',
                self.custom_urls, plugin=plugins.name(__file__))

        return False

    def load_streams(self, node, load_node, use_cache=True):
        stations = (
            (_('Personal'), 'lastfm://user/%s/personal' % self.config.username),
            (_('Recommended'), 'lastfm://user/%s/recommended/100' %
                self.config.username),
            (_('Neighbourhood'), 'lastfm://user/%s/neighbours' %
                self.config.username),
            (_('Loved Tracks'), 'lastfm://user/%s/loved' % self.config.username),
        )

        for station, url in stations:
            item = radio.RadioGenre(station, self)
            item.lastfm_url = url
            item.custom = False
            n = self.model.append(node, [self.note, item])
            item.node = n

        # custom
        for station, url in self.custom_urls:
            item = radio.RadioGenre(station, self)
            item.lastfm_url = url
            item.custom = True
            n = self.model.append(node, [self.note, item])
            item.node = n

        self.model.remove(load_node)
        self.last_node = node

    def handles_url(self, url):
        """
            Check to see if this is a lastfm:// url, and if so, return True
        """
        if re.match(r"^lastfm://.*$", url):
            return True

    def handle_url(self, url):
        """
            Handle a url
        """
        url = "lastfm://%s" % urllib.quote(url.replace('lastfm://', ''))
        item = radio.RadioGenre(url, self)
        item.lastfm_url = url
        self.load_genre(item)

    def load_genre(self, genre, rel=False):
        if not PROXY.proxy_ready:
            common.error(APP.window, _("The Last.FM Proxy has "
                "not yet connected to last.fm.  Please wait "
                "a moment and try again"))
            return
        current = self.exaile.player.current
        if hasattr(current, 'lastfm_track'):
            current.album = "LastFM: %s" % str(genre)
            new_loc = 'http://localhost:%d/lastfm.mp3' % self.listenport
            self.exaile.tracks.refresh_row(current)

            self.exaile.status.set_first(_("Changing stations..."), 3500)
            if not current.loc == new_loc:
                current.loc = new_loc
                self.exaile.player.play_track(current)
            if not current in self.exaile.tracks.songs:
                self.exaile.tracks.append_song(current)
            self.command("/changestation/%s" % genre.lastfm_url)
            xlmisc.log("LastFM: playing %s" % genre.lastfm_url)
            return
        tr = media.Track()
        tr.type = 'stream'
        tr.artist = _('LastFM Radio!')
        tr.album = "LastFM: %s" % str(genre)
        tr.title = "LastFM: %s" % str(genre)
        tr.loc = "http://localhost:%d/lastfm.mp3" % self.listenport
        tr.lastfm_track = True

        self.exaile.tracks.append_song(tr)
        self.exaile.player.play_track(tr)
        xlmisc.log("LastFM: playing %s" % genre.lastfm_url)
        self.command('/changestation/%s' % genre.lastfm_url)
        
    def __str__(self):
        return _("LastFM Radio")

def unzip_file(file, odir):
    """
        Unzips "file" to "odir" (output directory)
    """
    z = zipfile.ZipFile(file)
    for name in z.namelist():
        m = re.match(r'^(.*)/([^/]*)$', name)
        if m:
            dir = m.group(1)

            if not os.path.isdir(os.path.join(odir, dir)):
                os.makedirs(os.path.join(odir, dir))

            if not m.group(2): continue

        h = open(os.path.join(odir, name), 'w')
        h.write(z.read(name))
        h.close()

def load_data(zip):
    """
        Loads the data from the zipfile
    """
    global TMP_DIR, PLUGIN_ICON, GLADE_XML_STRING
    if TMP_DIR: return

    fname = os.sep.join(str(__file__).split(os.path.sep)[0:-1])
    TMP_DIR = "/tmp/lfmdir%s" % md5.new(str(random.randrange(0,
        123471287348834))).hexdigest()
    os.mkdir(TMP_DIR)
    unzip_file(fname, TMP_DIR)

    GLADE_XML_STRING = zip.get_data('data/lastfmproxy.glade')

@common.threaded
def run_proxy(config):
    RUN_COUNT = 0
    while True:
        try:
            PLUGIN.listenport = config.listenport
            PROXY.run(config.bind_address, config.listenport)
            return
        except Exception, e:
            if RUN_COUNT >= 5:
                raise(e)
            if e.args[0] == 98:
                RUN_COUNT += 1
                xlmisc.log("LastFM Proxy: Port %d in use, trying %d" %
                    (config.listenport, config.listenport + 1))
                config.listenport += 1
            else:
                raise(e)

BUTTON_ITEMS = (
    (_('LastFM: Skip this track'), 'gtk-media-forward', '/skip'),
    (_('LastFM: Mark this track as loved'), 'gtk-add', '/love'),
    (_('LastFM: Ban this track'), 'gtk-delete', '/ban'),
)

@common.threaded
def set_cover(value):
    if not OLD_COVER_FUNC or not hasattr(APP.player.current,
        'lastfm_track'): return
    fname = "/tmp/lfmcover%s" % md5.new(str(random.randrange(0,
        23423423442))).hexdigest()

    value = value.replace('130x130', '300x300')
    data = urllib.urlopen(value).read()
    h = open(fname, 'w+')
    h.write(data)
    h.close()
    APP.player.current._tmp_cover = fname

    gobject.idle_add(APP.cover.set_lfmimage, fname)

def album_callback(value):
    if OLD_COVER_FUNC:
        set_cover(value)
    xlmisc.log("LastFM: Got cover: %s" % value)

def new_cover_func(self, *e):
    """
        Do absolutely nothing.  Srs.
    """
    pass

SIGNAL_ID = None
def play_track(exaile, track):
    global SIGNAL_ID, OLD_COVER_FUNC
    if not SHOW_COVER or not hasattr(APP.player.current,
        'lastfm_track'): return
    
    OLD_COVER_FUNC = APP.cover.set_image
    APP.cover.set_image = new_cover_func
    APP.cover.set_lfmimage = OLD_COVER_FUNC

    APP.cover.set_lfmimage(xl.path.get_data('images', 'nocover.png'))
    SIGNAL_ID = APP.connect('timer_update', lambda *e:
        HTTP_CLIENT.req('/np'))

def stop_track(exaile, track):
    global SIGNAL_ID, OLD_COVER_FUNC
    if not SHOW_COVER: return
    if SIGNAL_ID:
        APP.disconnect(SIGNAL_ID)
        SIGNAL_ID = None
    if OLD_COVER_FUNC:
        APP.cover.set_lfmimage(xl.path.get_data('images', 'nocover.png'))
        APP.cover.set_image = OLD_COVER_FUNC
        OLD_COVER_FUNC = None

def initialize():
    global PROXY, PLUGIN, HTTP_CLIENT, BUTTONS, SHOW_COVER
    if not TMP_DIR in sys.path: sys.path.append(TMP_DIR)

    import lastfmmain
    import config
    import httpclient

    settings = APP.settings
    SHOW_COVER = settings.get_boolean('ui/show_cover', True)

    port = settings.get_int('listenport', plugin=plugins.name(__file__),
        default=1881)
    config.listenport = port
#    config.username = settings.get_str('lastfmuser',
#        plugin=plugins.name(__file__),
#        default=settings.get_str('lastfm/user', ''))
#    config.password = settings.get_crypted('lastfmpass',
#        plugin=plugins.name(__file__),
#        default=settings.get_str('lastfm/pass', ''))

#    if settings.get_boolean('lastfm_use_main', default=True,
#        plugin=plugins.name(__file__)):
    config.username = settings.get_str('lastfm/user', '')
    config.password = settings.get_crypted('lastfm/pass', '') 

    if not PLUGIN:
        PLUGIN = LastFMDriver(APP.pradio_panel)
        PLUGIN.config = config
        APP.pradio_panel.add_driver(PLUGIN, plugins.name(__file__))
        HTTP_CLIENT = httpclient.httpclient('localhost', config.listenport)
        APP.add_urlhandler(PLUGIN)

    PROXY = lastfmmain.proxy(config.username, config.password)
    PROXY.np_image_func = album_callback
    PROXY.basedir = TMP_DIR
    run_proxy(config)

    if not CONS.dict:
        CONS.connect(APP.player, 'play-track', play_track)
        CONS.connect(APP.player, 'stop-track', stop_track)

    if not BUTTONS:
        for tooltip, icon, command in BUTTON_ITEMS:
            button = gtk.Button()
            button.connect('clicked', lambda w, command=command: PLUGIN.command(command))
            image = gtk.Image()
            image.set_from_stock(icon, gtk.ICON_SIZE_MENU)
            button.set_size_request(32, 32)
            button.set_image(image)
            TIPS.set_tip(button, tooltip)
            APP.xml.get_widget('rating_toolbar').pack_start(button)
            BUTTONS.append(button)
        for button in BUTTONS:
            button.show()

    return True

def destroy():
    global PLUGIN, MENU_ITEM, PROXY, BUTTONS, OLD_COVER_FUNC
    if TMP_DIR:
        sys.path.remove(TMP_DIR) 

    CONS.disconnect_all()

    if OLD_COVER_FUNC:
        APP.cover.set_image = OLD_COVER_FUNC
        OLD_COVER_FUNC = None

    if PLUGIN:
        APP.pradio_panel.remove_driver(PLUGIN)
        APP.remove_urlhandler(PLUGIN)

    if BUTTONS:
        for button in BUTTONS:
            button.hide()
            button.destroy()

        BUTTONS = []
    
    if PROXY:
        PROXY.np_image_func = None
        PROXY.quit = True

    PROXY = None
    MENU_ITEM = None
    PLUGIN = None

#def use_main_toggled(box, user, password):
#    active = not box.get_active()
#    user.set_sensitive(active)
#    password.set_sensitive(active)

def quick_init():
    """
        Runs initialize, but returns False so the timer doesn't continue to
        run
    """
    global PROXY
    initialize()

    return False

def configure():
    global PROXY
    exaile = APP
    settings = exaile.settings

    xml = gtk.glade.xml_new_from_buffer(GLADE_XML_STRING,
        len(GLADE_XML_STRING))

    dialog = xml.get_widget('ConfigurationDialog')
#    use_main = xml.get_widget('lastfm_use_main')
#    lastfm_user = xml.get_widget('lastfm_user')
#    lastfm_pass = xml.get_widget('lastfm_pass')
    lastfm_listen_port = xml.get_widget('lastfm_listen_port')

#    use_main.set_active(settings.get_boolean('use_main',
#        plugin=plugins.name(__file__), default=True))

#    lastfm_user.set_text(settings.get_str('lastfmuser',
#        plugin=plugins.name(__file__),
#        default=settings.get_str('lastfm/user', '')))
#    lastfm_pass.set_text(settings.get_crypted('lastfmpass',
#        plugin=plugins.name(__file__),
#        default=settings.get_str('lastfm/pass', '')))

    lastfm_listen_port.set_text(settings.get_str('listenport',
        plugin=plugins.name(__file__), default=1881))

#    use_main.connect('toggled', lambda *e: use_main_toggled(use_main,
#        lastfm_user, lastfm_pass))
#    use_main_toggled(use_main, lastfm_user, lastfm_pass)

    result = dialog.run()
    dialog.hide()
    if result == gtk.RESPONSE_OK:
        if PROXY: 
            PROXY.np_image_func = None
            PROXY.quit = True
            PROXY.stop = True
            exaile.player.stop()
#        settings.set_boolean('use_main', use_main.get_active(),
#            plugin=plugins.name(__file__))
#        settings.set_str('lastfm_user', lastfm_user.get_text(), 
#            plugin=plugins.name(__file__))
#        settings.set_crypted('lastfm_pass', lastfm_pass.get_text(),
#            plugin=plugins.name(__file__))

        settings.set_str('listenport', lastfm_listen_port.get_text(),
            plugin=plugins.name(__file__))

        if PLUGIN_ENABLED:
            gobject.timeout_add(5000, quick_init)

icon_data = ["16 16 72 1",
" 	c None",
".	c #D20039",
"+	c #D71B4E",
"@	c #EE9EB4",
"#	c #F9DFE6",
"$	c #FAE5EB",
"%	c #F5C8D4",
"&	c #E05077",
"*	c #E04D75",
"=	c #F6CAD6",
"-	c #F7D3DD",
";	c #E87A98",
">	c #D71E50",
",	c #F8D8E1",
"'	c #F5C5D2",
")	c #E1547A",
"!	c #DD416B",
"~	c #EB8DA6",
"{	c #FDF5F7",
"]	c #E36084",
"^	c #D40A41",
"/	c #FAE0E7",
"(	c #E66F8F",
"_	c #E04F76",
":	c #EC94AC",
"<	c #D40D43",
"[	c #EFA2B7",
"}	c #F3BBCA",
"|	c #E25C80",
"1	c #DD3D68",
"2	c #DC3965",
"3	c #F8DAE2",
"4	c #E1577D",
"5	c #D3033B",
"6	c #F6CBD7",
"7	c #EA8AA4",
"8	c #E67191",
"9	c #FBEAEF",
"0	c #F3BAC9",
"a	c #E36184",
"b	c #D3083F",
"c	c #F9DCE4",
"d	c #E05279",
"e	c #FAE1E8",
"f	c #D40E44",
"g	c #D82253",
"h	c #E9829E",
"i	c #FBEBEF",
"j	c #ED9BB1",
"k	c #F0A9BC",
"l	c #F1B1C2",
"m	c #DB3562",
"n	c #F6CED9",
"o	c #E56B8C",
"p	c #E1557B",
"q	c #F9DDE5",
"r	c #D92657",
"s	c #FBE6EC",
"t	c #F2B4C5",
"u	c #DC3A66",
"v	c #D92959",
"w	c #E77997",
"x	c #FDF2F5",
"y	c #E15379",
"z	c #FBE8ED",
"A	c #DF4B73",
"B	c #DA2B5A",
"C	c #F2B8C8",
"D	c #D92A5A",
"E	c #FAE3E9",
"F	c #F7CFDA",
"G	c #E9839F",
" .............. ",
"................",
"................",
"................",
"..+@#$%&..*=-;..",
".>,')!~{]^/(_:<.",
".[}....|$1$2....",
".34....567890ab.",
".cd.....]efghij.",
".kl....mgno..pq.",
".rstuvwx2yzAB0C.",
"..Dl/EFa..G3clg.",
"................",
"................",
"................",
" .............. "]
PLUGIN_ICON = gtk.gdk.pixbuf_new_from_xpm_data(icon_data)



import gtk, gconf, gobject
import xl.plugins as plugins
from gettext import gettext as _

PLUGIN_NAME = _("Global Hotkeys")
PLUGIN_AUTHORS = ['Sayamindu Dasgupta <sayamindu@gnome.org>',
                  'Lincoln de Sousa <lincoln@archlinux-br.org']
PLUGIN_VERSION = '0.6.2'
PLUGIN_DESCRIPTION = _(r"""Enables support for Global Hotkeys in metacity
(default gnome window manager)""")
PLUGIN_ENABLED = False

b = gtk.Button()
PLUGIN_ICON = b.render_icon(gtk.STOCK_MEDIA_PLAY, gtk.ICON_SIZE_MENU)
b.destroy()

###
# Add keyboard shortcut items here in tuples in the following format:
#
#    setting name   shortcut    command         label
##
items = (
    ('show_hide',   '<Super>p', 'exaile',      _('Show/Hide Exaile:')),
    ('play',        '<Super>x', 'exaile -a',   _('Play:')),
    ('play_pause',  '<Super>c', 'exaile -t',   _('Toggle Play/Pause:')),
    ('stop',        '<Super>v', 'exaile -s',   _('Stop:')),
    ('previous',    '<Super>z', 'exaile -p',   _('Previous Song:')),
    ('next',        '<Super>b', 'exaile -n',   _('Next Song:')),
    ('ivolume',     '<Super>i', 'exaile -i 5', _('Increase Volume:')),
    ('dvolume',     '<Super>d', 'exaile -l 5', _('Decrease Volume:'))
)

# some shortcuts
_keyb = lambda cmd: '/apps/metacity/global_keybindings/run_command_%d' % cmd
_keyc = lambda cmd: '/apps/metacity/keybinding_commands/command_%d' % cmd
keyb = lambda c, cmd: c.get_string(_keyb(cmd))
keyc = lambda c, cmd: c.get_string(_keyc(cmd))
skeyb = lambda c, cmd, value: c.set_string(_keyb(cmd), value)
signal_id = None

def destroy():
    """
        Called when the plugin is disabled
    """
    global signal_id
    c = gconf.client_get_default()
    for i, v in enumerate(items):
        count = i + 20
        if keyc(c, count).find('exaile') > -1:
            c.unset(_keyb(count))
            c.unset(_keyc(count))

    if signal_id:
        gobject.source_remove(signal_id)
        signal_id = None

    return

def initialize():
    """
        Called when the plugin is enabled
    """
    global signal_id
    # TODO figure out if Metacity is the current window manager or not
    # Return False if metacity is not the current wm

    c = gconf.client_get_default()
    for i, v in enumerate(items):
        count = i + 20

        binding = APP.settings.get_str(v[0], default=v[1],
            plugin=plugins.name(__file__))

        # key binding
        c.set_string(_keyb(count), binding)

        # key command
        c.set_string(_keyc(count), v[2])

    signal_id = APP.connect('quit', lambda *e: destroy())

    return True

def configure():
    """
        Called when the user wants to configure the plugin
    """
    global t_count
    c = gconf.client_get_default()

    table = gtk.Table()
    table.set_border_width(8)
    fields = []
    t_count = 0
    
    def attach_field(label, value):
        global t_count
        label = gtk.Label(label)
        label.set_property('xalign', 1)
        label.set_padding(4, 4)

        input = gtk.Entry()
        input.set_text(value)
        table.attach(label, 0, 1, t_count, t_count + 1)
        table.attach(input, 1, 2, t_count, t_count + 1)
        fields.append(input)
        t_count += 1

    for i, v in enumerate(items):
        binding = APP.settings.get_str(v[0], default=v[1],
            plugin=plugins.name(__file__))
        attach_field(v[3], binding)

    frame = gtk.Frame(_('Set Hotkeys'))
    frame.set_border_width(8)
    frame.add(table)

    dlg = plugins.PluginConfigDialog(APP.window, PLUGIN_NAME)
    dlg.get_child().add(frame)
    dlg.show_all()
    response = dlg.run()
    dlg.hide()

    if response == gtk.RESPONSE_OK:
        for i, v in enumerate(items):
            if PLUGIN_ENABLED: skeyb(c, i + 20, fields[i].get_text())
            APP.settings.set_str(v[0], fields[i].get_text())



# Copyright (C) 2006 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 1, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import os, random, re, urllib
from gettext import gettext as _
import gtk
from xl import common, xlmisc, filtergui, library
from xl.filtergui import MultiEntryField, EntryField
import xl.path

N_ = lambda x: x

def day_calc(x, inc, field, symbol='>='):
    import time
    values = {
        'seconds': 1,
        'minutes': 60,
        'hours': 60 * 60,
        'days': 60 * 60 * 24,
        'weeks': 60 * 60 * 24 * 7
    }

    seconds = int(x) * values[inc]
    t = time.time() - seconds
    t = time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(t))
    return "%s %s '%s'" % (field, symbol, t)

class EntrySecondsField(MultiEntryField):
    def __init__(self, result_generator):
        MultiEntryField.__init__(self, result_generator, n=1,
            labels=(None, _('seconds')),
            widths=(50,))

class EntryAndEntryField(MultiEntryField):
    def __init__(self, result_generator):
        MultiEntryField.__init__(self, result_generator, n=2,
            # TRANSLATORS: Logical AND used for smart playlists
            labels=(None, _('and'), None),
            widths=(50, 50))

class EntryDaysField(MultiEntryField):
    def __init__(self, result_generator):
        MultiEntryField.__init__(self, result_generator, n=1,
            labels=(None, _('days')),
            widths=(50,))

DATE_FIELDS = (_('seconds'), _('minutes'), _('hours'), _('days'), _('weeks'))
class SpinDateField(filtergui.SpinButtonAndComboField):
    def __init__(self, result_generator):
        filtergui.SpinButtonAndComboField.__init__(self, 
            result_generator, DATE_FIELDS)

class SpinSecondsField(filtergui.SpinLabelField):
    def __init__(self, result_generator):
        filtergui.SpinLabelField.__init__(self, result_generator, 
            _('seconds'))

class SpinRating(filtergui.SpinLabelField):
    def __init__(self, result_generator):
        filtergui.SpinLabelField.__init__(self, result_generator, '',
            8, -8)

class SpinNothing(filtergui.SpinLabelField):
    def __init__(self, result_generator):
        filtergui.SpinLabelField.__init__(self, result_generator, '')

CRITERIA = [
    (N_('Artist'), [
        # TRANSLATORS: True if haystack is equal to needle
        (N_('is'), (EntryField, lambda x:
            'artists.name = "%s"' % x)),
        # TRANSLATORS: True if haystack is not equal to needle
        (N_('is not'), (EntryField, lambda x:
            'artists.name != "%s"' % x)),
        # TRANSLATORS: True if haystack contains needle
        (N_('contains'), (EntryField, lambda x:
            'artists.name LIKE "%%%s%%"' % x)),
        # TRANSLATORS: True if haystack does not contain needle
        (N_('does not contain'), (EntryField, lambda x:
            'artists.name NOT LIKE "%%%s%%"' % x)),
        ]),
    (N_('Album'), [
        (N_('is'), (EntryField, lambda x:
            'albums.name = "%s"' % x)),
        (N_('is not'), (EntryField, lambda x:
            'albums.name != "%s"' % x)),
        (N_('contains'), (EntryField, lambda x:
            'albums.name LIKE "%%%s%%"' % x)),
        (N_('does not contain'), (EntryField, lambda x:
            'albums.name NOT LIKE "%%%s%%"' % x)),
        ]),
    (N_('Genre'), [
        (N_('is'), (EntryField, lambda x:
            'genre = "%s"' % x)),
        (N_('is not'), (EntryField, lambda x:
            'genre != "%s"' % x)),
        (N_('contains'), (EntryField, lambda x:
            'genre LIKE "%%%s%%"' % x)),
        (N_('does not contain'), (EntryField, lambda x:
            'genre NOT LIKE "%%%s%%"' %x)),
        ]),
    (N_('User Rating'), [
        # TRANSLATORS: Example: rating >= 5
        (N_('at least'), (SpinRating, lambda x:
            'user_rating >= %s' % x)),
        # TRANSLATORS: Example: rating <= 3
        (N_('at most'), (SpinRating, lambda x:
            'user_rating <= %s' % x))]),
    (N_('System Rating'), [
        (N_('at least'), (SpinRating, lambda x:
            'rating >= %s' % x)),
        (N_('at most'), (SpinRating, lambda x:
            'rating <= %s' % x))
        ]),
    (N_('Number of Plays'), [
        (N_('at least'), (SpinNothing, lambda x:
            'plays >= %s' % x)),
        (N_('at most'), (SpinNothing, lambda x:
            'plays <= %s' %x))
        ]),
    (N_('Year'), [
        # TRANSLATORS: Example: year < 1999
        (N_('before'), (EntryField, lambda x:
            'year < %s' % x)),
        # TRANSLATORS: Example: year > 2002
        (N_('after'), (EntryField, lambda x:
            'year > %s' % x)),
        # TRANSLATORS: Example: 1980 <= year <= 1987
        (N_('between'), (EntryAndEntryField, lambda x, y:
            'year BETWEEN %s AND %s' % (x, y))),
        ]),
    (N_('Length'), [
        (N_('at least'), (SpinSecondsField, lambda x:
            'length >= %s' % x)),
        (N_('at most'), (SpinSecondsField, lambda x:
            'length <= %s' % x)),
        ]),
    (N_('Date Added'), [
        # TRANSLATORS: Example: track has been added in the last 2 days
        (N_('in the last'), (SpinDateField, 
            lambda x, i: day_calc(x, i, 'time_added'))),
        # TRANSLATORS: Example: track has not been added in the last 5 hours
        (N_('not in the last'), (SpinDateField, 
            lambda x, i: day_calc(x, i, 'time_added', '<'))),
        ]),
    (N_('Last Played'), [
        (N_('in the last'), (SpinDateField, 
            lambda x, i: day_calc(x, i, 'last_played'))),
        (N_('not in the last'), (SpinDateField, 
            lambda x, i: day_calc(x, i, 'last_played', '<'))),
        ]),
    (N_('Location'), [
        (N_('is'), (EntryField, lambda x:
            'paths.name = "%s"' % x)),
        (N_('is not'), (EntryField, lambda x:
            'paths.name != "%s"' % x)),
        (N_('contains'), (EntryField, lambda x:
            'paths.name LIKE "%%%s%%"' % x)),
        (N_('does not contain'), (EntryField, lambda x:
            'paths.name NOT LIKE "%%%s%%"' % x)),
        ])
    ]


class SmartPlaylist(object):
    def __init__(self, name, id):
        self.name = name
        self.id = id

    def __str__(self):
        return self.name

class BuiltinPlaylist(object):
    def __init__(self, name, sql):
        self.name = name
        self.sql = sql

    def __str__(self):
        return self.name

class CustomPlaylist(object):
    def __init__(self, name, id):
        self.name = name
        self.id = id

    def __str__(self):
        return self.name

class PlaylistsPanel(object):
    """ 
        The playlists panel 
    """
    def __init__(self, exaile):
        """
            Creates the playlist panel
        """
        self.exaile = exaile
        self.db = self.exaile.db
        self.xml = exaile.xml
        container = self.xml.get_widget('playlists_box')

        self.targets = [('text/uri-list', 0, 0)]
        self.tree = xlmisc.DragTreeView(self, True, False)
        self.tree.connect('row-activated', self.open_playlist)
        self.tree.set_headers_visible(False)

        self.scroll = gtk.ScrolledWindow()
        self.scroll.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        self.scroll.add(self.tree)
        self.scroll.set_shadow_type(gtk.SHADOW_IN)
        container.pack_start(self.scroll, True, True)
        container.show_all()

        pb = gtk.CellRendererPixbuf()
        cell = gtk.CellRendererText()
        col = gtk.TreeViewColumn('Text')
        col.pack_start(pb, False)
        col.pack_start(cell, True)
        col.set_attributes(pb, pixbuf=0)
        col.set_attributes(cell, text=1)
        self.tree.append_column(col)

    def load_playlists(self):
        """
            Loads the playlists
        """
        self.model = gtk.TreeStore(gtk.gdk.Pixbuf, str, object)
        self.tree.set_model(self.model)
        self.open_folder = xlmisc.get_icon('gnome-fs-directory-accept')
        self.playlist_image = gtk.gdk.pixbuf_new_from_file(xl.path.get_data(
            'images', 'playlist.png'))
        self.smart_image = self.exaile.window.render_icon('gtk-execute',
            gtk.ICON_SIZE_MENU)
        self.smart = self.model.append(None, [self.open_folder, _("Smart"
            " Playlists"), None])

        smart_playlists = [
            (_('Entire Library'), "SELECT paths.name FROM artists, albums, tracks, paths WHERE " \
                "paths.id=tracks.path AND artists.id=tracks.artist AND " \
                "albums.id=tracks.album ORDER BY LOWER(artists.name), " \
                "THE_CUTTER(albums.name)"),

            (_('Top 100'), "SELECT paths.name FROM tracks,paths WHERE " \
                "paths.id=tracks.path ORDER BY rating " \
                "DESC LIMIT 100"),

            (_('Highest Rated'), "SELECT paths.name FROM tracks,paths WHERE " \
                "tracks.path=paths.id " \
                "ORDER BY user_rating DESC " \
                "LIMIT 100"),

            (_('Most Played'), "SELECT paths.name FROM paths, tracks WHERE " \
                "tracks.path=paths.id ORDER " \
                "BY plays DESC LIMIT 100"),

            (_('Least Played'), "SELECT paths.name FROM paths, tracks WHERE " \
                "paths.id=tracks.path ORDER " \
                "BY plays ASC LIMIT 100"),

            (_('Rating > 4'), "SELECT paths.name FROM paths, tracks, artists, " \
                "albums " \
                "WHERE tracks.path=paths.id AND albums.id=tracks.album AND " \
                "artists.id=tracks.artist " \
                "AND user_rating > 4 " \
                "ORDER BY LOWER(artists.name), THE_CUTTER(albums.name), track"),
            
            (_('Rating > 3'), "SELECT paths.name FROM paths,tracks,artists,albums WHERE " \
                "tracks.path=paths.id AND albums.id=tracks.album AND " \
                "artists.id=tracks.artist AND user_rating > 3 " \
                "ORDER BY LOWER(artists.name), LOWER(albums.name), track"),

            (_('Newest 100'), "SELECT paths.name FROM paths,tracks WHERE " \
                "tracks.path=paths.id AND time_added!='' " \
                "ORDER BY time_added DESC " \
                "LIMIT 100"),

            (_('Random 100'), "SELECT paths.name FROM tracks,paths WHERE " \
                "paths.id=tracks.path ORDER BY RANDOM() LIMIT 100"),
            (_('Random 500'), "SELECT paths.name FROM tracks,paths WHERE " \
                "paths.id=tracks.path ORDER BY RANDOM() LIMIT 500"),
        ]

        # Add smart playlists
        def add_smart(parent, i):
            name, sql = smart_playlists[i]
            self.model.append(parent, [self.playlist_image, name,
                BuiltinPlaylist(name, sql)])

        add_smart(self.smart, 0)
        builtin = self.model.append(self.smart, [self.open_folder,
            _('Built In'), None])
        for i in xrange(1, len(smart_playlists)):
            add_smart(builtin, i)

        self.custom = self.model.append(None, [self.open_folder,
            _("Custom Playlists"), None])

        rows = self.db.select("SELECT name, id, type FROM playlists ORDER BY"
            " name")
        for row in rows:
            if not row[2]:
                self.model.append(self.custom, [self.playlist_image, row[0],
                    CustomPlaylist(row[0], row[1])])
            elif row[2] == 1:
                self.model.append(self.smart, [self.smart_image, row[0],
                    SmartPlaylist(row[0], row[1])])

        self.tree.expand_all()
        self.setup_menu()

    def setup_menu(self):
        """ 
            Sets up the popup menu for the playlist tree
        """
        self.menu = xlmisc.Menu()
        self.menu.append(_('Add Playlist'), self.on_add_playlist, 'gtk-add')
        self.menu.append(_('Add Smart Playlist'), self.on_add_smart_playlist, 
            'gtk-add')
        self.menu.append_separator()
        self.edit_item = self.menu.append(_('Edit'), self.edit_playlist,
            'gtk-edit')
        self.menu.append_separator()
        self.remove_item = self.menu.append(_('Delete Playlist'), 
            self.remove_playlist, 'gtk-remove')

    def edit_playlist(self, item, event):
        """
            Edits a playlist
        """
        selection = self.tree.get_selection()
        (model, iter) = selection.get_selected()

        obj = model.get_value(iter, 2)
        if not isinstance(obj, SmartPlaylist): return
        row = self.db.read_one('playlists', 'matchany, item_limit', 
            'id=?', (obj.id,))

        dialog = filtergui.FilterDialog(_('Edit Playlist'), CRITERIA)
        dialog.set_transient_for(self.exaile.window)

        dialog.set_name(obj.name)
        dialog.set_match_any(row[0])
        dialog.set_limit(row[1])

        state = []
        rows = self.db.select('SELECT crit1, crit2, filter FROM '
            'smart_playlist_items WHERE playlist=? ORDER BY line',
            (obj.id,))

        for row in rows:
            left = [row[0], row[1]]
            filter = eval(row[2])
            if len(filter) == 1:
                filter = filter[0]
            state.append((left, filter))

        print repr(state)

        dialog.set_state(state)

        result = dialog.run()
        dialog.hide()
        if result == gtk.RESPONSE_ACCEPT:
            name = dialog.get_name()
            if name != obj.name:
                row = self.db.read_one('playlists', 'matchany', 
                    'name=?', (name,))
                if row:
                    common.error(self.exaile.window, _("That playlist name "
                        "is already taken."))
                    return
            matchany = dialog.get_match_any()
            limit = dialog.get_limit()
            self.db.execute('UPDATE playlists SET name=?, matchany=?, item_limit=?WHERE '
                'id=?', (name, matchany, limit, obj.id))
            self.db.execute('DELETE FROM smart_playlist_items WHERE '
                'playlist=?', (obj.id,))

            count = 0
            for c, v in dialog.get_state():
                if type(v) != list:
                    v = list((v,))
                self.db.execute("INSERT INTO smart_playlist_items( "
                    "playlist, line, crit1, crit2, filter ) VALUES( "
                    " ?, ?, ?, ?, ? )", (obj.id, count, c[0], c[1],
                    repr(v)))
                count += 1

            self.db.commit()
            self.model.set_value(iter, 1, name)
            self.model.set_value(iter, 2, SmartPlaylist(name, obj.id))

        dialog.destroy()

    def on_add_smart_playlist(self, widget, event):
        """
            Adds a smart playlist
        """
        dialog = filtergui.FilterDialog(_('Add Smart Playlist'), CRITERIA)

        dialog.set_transient_for(self.exaile.window)
        result = dialog.run()
        if result == gtk.RESPONSE_ACCEPT:
            name = dialog.get_name()
            matchany = dialog.get_match_any()
            limit = dialog.get_limit()
            if not name: 
                common.error(self.exaile.window, _("You did not enter a "
                    "name for your playlist"))
                return
            row = self.db.read_one('playlists', 'name', 'name=?', (name,))
            if row:
                common.error(self.exaile.window, _("That playlist name "
                    "is already taken."))
                return
            dialog.hide()

            self.db.execute("INSERT INTO playlists( name, type, matchany, "
                " item_limit ) VALUES( ?, 1, ?, ? )", (name, matchany, limit))
            row = self.db.read_one('playlists', 'id', 'name=?', (name,))
            playlist_id = row[0]

            count = 0
            for c, v in dialog.get_state():
                if type(v) != list:
                    v = list((v,))
                self.db.execute("INSERT INTO smart_playlist_items( "
                    "playlist, line, crit1, crit2, filter ) VALUES( "
                    " ?, ?, ?, ?, ? )", (playlist_id, count, c[0], c[1],
                    repr(v)))
                count += 1

            self.db.commit()

            self.model.append(self.smart, [self.smart_image, name, 
                SmartPlaylist(name, playlist_id)])

        dialog.hide()
        dialog.destroy()

    def button_press(self, widget, event):
        """
            Called when the user clicks on the tree
        """
        selection = self.tree.get_selection()
        x, y = event.get_coords()
        x = int(x); y = int(y)
        delete_enabled = False
        edit_enabled = False
        if self.tree.get_path_at_pos(x, y):
            (path, col, x, y) = self.tree.get_path_at_pos(x, y)
            iter = self.model.get_iter(path)
            obj = self.model.get_value(iter, 2)
            self.edit_item.set_sensitive(edit_enabled)
            if isinstance(obj, CustomPlaylist) or \
                isinstance(obj, SmartPlaylist):
                delete_enabled = True
            if isinstance(obj, SmartPlaylist):
                edit_enabled = True

        self.remove_item.set_sensitive(delete_enabled)
        self.edit_item.set_sensitive(edit_enabled)

        if event.button == 3:
            self.menu.popup(None, None, None, event.button, event.time)

    def open_playlist(self, tree, path, col):
        """
            Called when the user double clicks on a tree item
        """
        iter = self.model.get_iter(path)
        obj = self.model.get_value(iter, 2)
        if isinstance(obj, BuiltinPlaylist):
            name = obj.name
            sql = obj.sql

            if name == 'Entire Library':    
                songs = self.exaile.all_songs
            else:
                songs = library.search_tracks(self.exaile.window, 
                    self.db,
                    self.exaile.all_songs, None, None, sql)

            self.exaile.new_page(name, songs)

        elif isinstance(obj, SmartPlaylist):
            self.open_smart_playlist(obj.name, obj.id)

        elif isinstance(obj, CustomPlaylist):
            playlist = obj.name
            playlist_id = library.get_column_id(self.db, 'playlists', 'name', playlist)

            rows = self.db.select('SELECT paths.name FROM playlist_items,paths '
                'WHERE playlist_items.path=paths.id AND playlist=? ORDER BY playlist_items.sort_index',
                (playlist_id,))

            songs = library.TrackData()
            for row in rows:
                tr = library.read_track(self.db, self.exaile.all_songs, row[0])
                if tr:
                    songs.append(tr)

            self.playlist_songs = songs
            self.exaile.new_page(playlist, self.playlist_songs)
            self.exaile.on_search()
            self.exaile.tracks.playlist = playlist

    def open_smart_playlist(self, name, id):
        """
            Opens a smart playlist
        """
        row = self.db.read_one('playlists', 'matchany, item_limit', 'id=?', (id,))
        limit = row[1]
        rows = self.db.select("SELECT crit1, crit2, filter FROM "
            "smart_playlist_items WHERE playlist=? ORDER BY line", (id,))

        where = []
        andor = " AND "
        if row[0]: andor = ' OR '

        state = []

        for row in rows:
            left = [row[0], row[1]]
            filter = eval(row[2])
            if len(filter) == 1:
                filter = filter[0]
            state.append((left, filter))

        filter = filtergui.FilterWidget(CRITERIA)
        filter.set_state(state)
        where = filter.get_result()

        sql = """
            SELECT paths.name 
                FROM tracks,paths,artists,albums 
            WHERE 
                (
                    paths.id=tracks.path AND 
                    artists.id=tracks.artist AND 
                    albums.id=tracks.album
                ) 
                AND (%s) 
                ORDER BY 
                    LOWER(artists.name),
                    THE_CUTTER(albums.name), 
                    track, title
            """ % andor.join(where)
        xlmisc.log(sql)
        songs = library.search_tracks(self.exaile.window,
            self.db, self.exaile.all_songs, None, None, sql)

        # if a limit was set, shuffle the songs and only grab that amount
        if limit:
            random.shuffle(songs)
            songs = library.TrackData(songs[:limit])

        self.exaile.new_page(name, songs)

    def drag_data_received(self, tv, context, x, y, selection, info, etime):
        """
            Called when someone drags tracks to the smart playlists panel
        """
        path = self.tree.get_path_at_pos(x, y)
        error = ""
        if path: 
            iter = self.model.get_iter(path[0])
            obj = self.model.get_value(iter, 2)
        else:
            obj = None

        uris = selection.get_uris()
        songs = library.TrackData()
        for l in uris:
            l = urllib.unquote(l)
            m = re.search(r'^device_(\w+)://', l)
            if m:
                continue
            else:
                song = self.exaile.all_songs.for_path(l)
                if song: songs.append(song)

        if not isinstance(obj, CustomPlaylist): 
            self.on_add_playlist(self.remove_item, items=songs)
            return

        xlmisc.log("Adding tracks to playlist %s" % obj.name)
        if error:
            common.scrolledMessageDialog(self.exaile.window,
                error, _("The following errors did occur"))

        self.add_items_to_playlist(obj.name, songs)

    def remove_playlist(self, item, event):
        """
            Asks if the user really wants to delete the selected playlist, and
            then does so if they choose 'Yes'
        """
        selection = self.tree.get_selection()
        (model, iter) = selection.get_selected()

        obj = model.get_value(iter, 2)
        if not isinstance(obj, CustomPlaylist) and \
            not isinstance(obj, SmartPlaylist): return

        dialog = gtk.MessageDialog(self.exaile.window, 
            gtk.DIALOG_MODAL, gtk.MESSAGE_QUESTION, gtk.BUTTONS_YES_NO, 
            _("Are you sure you want to permanently delete the selected"
            " playlist?"))
        if dialog.run() == gtk.RESPONSE_YES:
            playlist = obj.name 
            p_id = library.get_column_id(self.db, 'playlists', 'name', playlist)
            self.db.execute("DELETE FROM playlists WHERE id=?", (p_id,))

            table = 'playlist_items'
            if isinstance(obj, SmartPlaylist):
                table = 'smart_playlist_items'
            self.db.execute("DELETE FROM %s WHERE playlist=?" % table,
                (p_id,))
            if library.PLAYLISTS.has_key(playlist):
                del library.PLAYLISTS[playlist]
            self.db.commit()
            
            self.model.remove(iter)
        dialog.destroy()

    def on_add_playlist(self, widget, event=None, items=None):
        """
            Adds a playlist to the database
        """
        dialog = common.TextEntryDialog(self.exaile.window, 
            _("Enter the name you want for your new playlist"),
            _("New Playlist"))
        result = dialog.run()
        if result == gtk.RESPONSE_OK:
            name = dialog.get_value()
            if name == "": return None
            c = self.db.record_count("playlists", "name=?", (name,))

            if c > 0:
                common.error(self.exaile.window, _("Playlist already exists."))
                return name

            playlist_id = library.get_column_id(self.db, 'playlists', 'name',
                name)
                
            self.model.append(self.custom, [self.playlist_image, name,
                CustomPlaylist(name, playlist_id)])
            self.tree.expand_all()

            if type(widget) == gtk.MenuItem:
                self.add_items_to_playlist(name, items)
            return name
        else: return None

    def add_items_to_playlist(self, playlist, songs=None):
        """
            Adds the selected tracks to the playlist
        """
        if type(playlist) == gtk.MenuItem:
            playlist = playlist.get_child().get_label()

        if songs == None: songs = self.exaile.tracks.get_selected_tracks()
        playlist_id = library.get_column_id(self.db, 'playlists', 'name', playlist)

        for track in songs:
            if track.type == 'stream': continue
            path_id = library.get_column_id(self.db, 'paths', 'name', track.loc)
            self.db.execute("INSERT INTO playlist_items( playlist, path ) " \
                "VALUES( ?, ? )", (playlist_id, path_id))
        self.db.commit()




#!/usr/bin/env python

# Copyright (C) 2006 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 1, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import time, subprocess

from urllib import pathname2url

def launch_burner(program, songs):
    """
        Launches a burner with the specified songs as options.
    """
    if not songs or not program:
        return 0

    # serpentine and brasero can be launched very similarly, k3b needs
    # special handling
    if program == 'serpentine':
        args = ['serpentine', '-o']
    if program == 'brasero':
        args = ['brasero', '-a']
        songs.reverse()
    if program == 'k3b':
        launch_k3b(songs)
        return

    ar = ['file://%s' % (pathname2url(song.io_loc),) \
            for song in songs if not song.type == 'stream']
    if not ar: return
    args.extend(ar)
    subprocess.Popen(args, stdout=-1,
        stderr=-1)

def check_burn_progs():
    """
        Function to check which of the supported burning programs
        are present.
    """

    # this function get called quite often, perhaps just check it once
    # when the program starts and then use the results throughout the
    # whole program?
    try:
        found = [program for program in ('serpentine', 'brasero', 'k3b') \
                if subprocess.call(['which', program], stdout=-1) == 0]
    except OSError: # No "which"
        found = []

    return found

def launch_k3b(songs):
    """
        Start k3b and create an audio project using (command line) DCOP
    """
    subprocess.call(["k3b", "--audiocd"])

    time.sleep(5) # k3b crashes if it's starting up and we query dcop right 
                  # away (stupid qt programs :P)

    # equivalent to the shell's  PROJECT=`dcop k3b K3bInterface currentProject`
    project = subprocess.Popen(['dcop', 'k3b', 'K3bInterface', 'currentProject'], \
            stdout=subprocess.PIPE).communicate()[0]
    project = project.strip()

    # addUrls takes a list of the form '[ song1 song2 ... ]'
    args = ['dcop', project, 'addUrls', '[']
    args.extend([song.io_loc for song in songs])
    args.append(']')
    subprocess.call(args)



"""
A pure-python library to assist sending data to AudioScrobbler (the LastFM
backend)
"""
import urllib, urllib2
from time import mktime
from datetime import datetime, timedelta
from md5 import md5
from xl import xlmisc

SESSION_ID = None
POST_URL   = None
NOW_URL    = None
HARD_FAILS = 0
LAST_HS    = None   # Last handshake time
HS_DELAY   = 0      # wait this many seconds until next handshake
SUBMIT_CACHE = []
MAX_CACHE  = 5      # keep only this many songs in the cache
PROTOCOL_VERSION = '1.2'
__LOGIN      = {}     # data required to login

class BackendError(Exception):
   "Raised if the AS backend does something funny"
   pass
class AuthError(Exception):
   "Raised on authencitation errors"
   pass
class PostError(Exception):
   "Raised if something goes wrong when posting data to AS"
   pass
class SessionError(Exception):
   "Raised when problems with the session exist"
   pass
class ProtocolError(Exception):
   "Raised on general Protocol errors"
   pass

def login( user, password, hashpw=False, client=('tst', '1.0') ):
   """Authencitate with AS (The Handshake)

   @param user:     The username
   @param password: md5-hash of the user-password
   @param hashpw:   If True, then the md5-hash of the password is performed
                    internally. If set to False (the default), then the module
                    assumes the passed value is already a valid md5-hash of the
                    password.
   @param client:   Client information (see http://www.audioscrobbler.net/development/protocol/ for more info)
   @type  client:   Tuple: (client-id, client-version)"""
   global LAST_HS, SESSION_ID, POST_URL, NOW_URL, HARD_FAILS, HS_DELAY, \
          PROTOCOL_VERSION, __LOGIN

   __LOGIN['u'] = user
   __LOGIN['c'] = client

   if LAST_HS is not None:
      next_allowed_hs = LAST_HS + timedelta(seconds=HS_DELAY)
      if datetime.now() < next_allowed_hs:
         delta = next_allowed_hs - datetime.now()
         raise ProtocolError("""Please wait another %d seconds until next handshake
(login) attempt.""" % delta.seconds)

   LAST_HS = datetime.now()

   tstamp = int(mktime(datetime.now().timetuple()))
   url    = "http://post.audioscrobbler.com/"

   if hashpw is True:
      __LOGIN['p'] = md5(password).hexdigest()
   else:
      __LOGIN['p'] = password

   token  = md5( "%s%d" % (__LOGIN['p'], int(tstamp))).hexdigest()
   values = {
         'hs': 'true',
         'p' : PROTOCOL_VERSION,
         'c': client[0],
         'v': client[1],
         'u': user,
         't': tstamp,
         'a': token
         }
   data = urllib.urlencode(values)
   req = urllib2.Request("%s?%s" % (url, data) )
   response = urllib2.urlopen(req)
   result = response.read()
   lines = result.split('\n')


   if lines[0] == 'BADAUTH':
      raise AuthError('Bad username/password')

   elif lines[0] == 'BANNED':
      raise Exception('''This client-version was banned by Audioscrobbler. Please
contact the author of this module!''')

   elif lines[0] == 'BADTIME':
      raise ValueError('''Your system time is out of sync with Audioscrobbler.
Consider using an NTP-client to keep you system time in sync.''')

   elif lines[0].startswith('FAILED'):
      handle_hard_error()
      raise BackendError("Authencitation with AS failed. Reason: %s" %
            lines[0])

   elif lines[0] == 'OK':
      # wooooooohooooooo. We made it!
      SESSION_ID = lines[1]
      NOW_URL    = lines[2]
      POST_URL   = lines[3]
      HARD_FAILS = 0
      xlmisc.log("[Last.FM]: Logged in successfully")

   else:
      # some hard error
      handle_hard_error()

def handle_hard_error():
   "Handles hard errors."
   global SESSION_ID, HARD_FAILS, HS_DELAY

   if HS_DELAY == 0:
      HS_DELAY = 60
   elif HS_DELAY < 120*60:
      HS_DELAY *= 2
   if HS_DELAY > 120*60:
      HS_DELAY = 120*60

   HARD_FAILS += 1
   if HARD_FAILS == 3:
      SESSION_ID = None

def now_playing( artist, track, album="", length="", trackno="", mbid="",
    inner_call=False ):
   """Tells audioscrobbler what is currently running in your player. This won't
   affect the user-profile on last.fm. To do submissions, use the "submit"
   method

   @param artist:  The artist name
   @param track:   The track name
   @param album:   The album name
   @param length:  The song length in seconds
   @param trackno: The track number
   @param mbid:    The MusicBrainz Track ID
   @return: True on success, False on failure"""

   global SESSION_ID, NOW_URL

   if SESSION_ID is None:
      raise AuthError("Please 'login()' first. (No session available)")

   if POST_URL is None:
      raise PostError("Unable to post data. Post URL was empty!")

   if length != "" and type(length) != type(1):
      raise TypeError("length should be of type int")

   if trackno != "" and type(trackno) != type(1):
      raise TypeError("trackno should be of type int")

   values = {'s': SESSION_ID,
             'a': unicode(artist).encode('utf-8'),
             't': unicode(track).encode('utf-8'),
             'b': unicode(album).encode('utf-8'),
             'l': length,
             'n': trackno,
             'm': mbid }

   data = urllib.urlencode(values)

   req = urllib2.Request(NOW_URL, data)
   response = urllib2.urlopen(req)
   result = response.read()

   if result.strip() == "OK":
      xlmisc.log("[Last.FM]: Submitted 'now playing' succesfully")
      return True
   elif result.strip() == "BADSESSION" :
      if inner_call is False:
         login(__LOGIN['u'], __LOGIN['p'], client=__LOGIN['c'])
         now_playing(artist, track, album, length, trackno, mbid, inner_call=True)
      else:
         raise SessionError('Invalid session')
   else:
      xlmisc.log("[Last.FM]: Error submitting 'now playing'")
      return False

def submit(artist, track, time, source='P', rating="", length="", album="",
      trackno="", mbid="", autoflush=False):
   """Append a song to the submission cache. Use 'flush()' to send the cache to
   AS. You can also set "autoflush" to True.

   From the Audioscrobbler protocol docs:
   ---------------------------------------------------------------------------

   The client should monitor the user's interaction with the music playing
   service to whatever extent the service allows. In order to qualify for
   submission all of the following criteria must be met:

   1. The track must be submitted once it has finished playing. Whether it has
      finished playing naturally or has been manually stopped by the user is
      irrelevant.
   2. The track must have been played for a duration of at least 240 seconds or
      half the track's total length, whichever comes first. Skipping or pausing
      the track is irrelevant as long as the appropriate amount has been played.
   3. The total playback time for the track must be more than 30 seconds. Do
      not submit tracks shorter than this.
   4. Unless the client has been specially configured, it should not attempt to
      interpret filename information to obtain metadata instead of tags (ID3,
      etc).

   @param artist: Artist name
   @param track:  Track name
   @param time:   Time the track *started* playing in the UTC timezone (see
                  datetime.utcnow()).

                  Example: int(time.mktime(datetime.utcnow()))
   @param source: Source of the track. One of:
                  'P': Chosen by the user
                  'R': Non-personalised broadcast (e.g. Shoutcast, BBC Radio 1)
                  'E': Personalised recommendation except Last.fm (e.g.
                       Pandora, Launchcast)
                  'L': Last.fm (any mode). In this case, the 5-digit Last.fm
                       recommendation key must be appended to this source ID to
                       prove the validity of the submission (for example,
                       "L1b48a").
                  'U': Source unknown
   @param rating: The rating of the song. One of:
                  'L': Love (on any mode if the user has manually loved the
                       track)
                  'B': Ban (only if source=L)
                  'S': Skip (only if source=L)
                  '':  Not applicable
   @param length: The song length in seconds
   @param album:  The album name
   @param trackno:The track number
   @param mbid:   MusicBrainz Track ID
   @param autoflush: Automatically flush the cache to AS?
   @return:       True on success, False if something went wrong
   """

   global SUBMIT_CACHE, MAX_CACHE

   source = source.upper()
   rating = rating.upper()

   if source == 'L' and (rating == 'B' or rating == 'S'):
      raise ProtocolError("""You can only use rating 'B' or 'S' on source 'L'.
See the docs!""")

   if source == 'P' and length == '':
      raise ProtocolError("""Song length must be specified when using 'P' as
source!""")

   if type(time) != type(1):
      raise ValueError("""The time parameter must be of type int (unix
timestamp). Instead it was %s""" % time)

   SUBMIT_CACHE.append(
         { 'a': unicode(artist).encode('utf-8'),
           't': unicode(track).encode('utf-8'),
           'i': time,
           'o': source,
           'r': rating,
           'l': length,
           'b': unicode(album).encode('utf-8'),
           'n': trackno,
           'm': mbid
            }
         )

   if autoflush or len(SUBMIT_CACHE) >= MAX_CACHE:
      return flush()
   else:
      return True

def flush(inner_call=False):
   """Sends the cached songs to AS.

   @param inner_call: Internally used variable. Don't touch!"""
   global SUBMIT_CACHE, __LOGIN

   if POST_URL is None:
      raise ProtocolError('''Cannot submit without having a valid post-URL. Did
you login?''')

   values = {}

   for i, item in enumerate(SUBMIT_CACHE):
      for key in item:
         values[key + "[%d]" % i] = item[key]

   values['s'] = SESSION_ID

   data = urllib.urlencode(values)
   req = urllib2.Request(POST_URL, data)
   response = urllib2.urlopen(req)
   result = response.read()
   lines = result.split('\n')

   if lines[0] == "OK":
      SUBMIT_CACHE = []
      xlmisc.log("[Last.FM]: OK: %s" % data)
      return True
   elif lines[0] == "BADSESSION" :
      if inner_call is False:
         login(__LOGIN['u'], __LOGIN['p'], client=__LOGIN['c'])
         flush(inner_call=True)
      else:
         raise Warning("Infinite loop prevented")
   elif lines[0].startswith('FAILED'):
      handle_hard_error()
      raise BackendError("Submission to AS failed. Reason: %s" %
            lines[0])
   else:
      # some hard error
      handle_hard_error()
      return False

if __name__ == "__main__":
   login( 'user', 'password' )
   submit(
         'De/Vision',
         'Scars',
         1192374052,
         source='P',
         length=3*60+44
         )
   submit(
         'Spineshank',
         'Beginning of the End',
         1192374052+(5*60),
         source='P',
         length=3*60+32
         )
   submit(
         'Dry Cell',
         'Body Crumbles',
         1192374052+(10*60),
         source='P',
         length=3*60+3
         )
   print flush()




#!/usr/bin/env python
# -*- coding: utf-8 -*-

import dbus, gobject, gtk, os
from gettext import gettext as _
import xl.plugins as plugins


# Plugin description
PLUGIN_NAME        = _('IM Status')
PLUGIN_ICON        = None
PLUGIN_ENABLED     = False
PLUGIN_AUTHORS     = ['Ingelrest François <Athropos@gmail.com>']
PLUGIN_VERSION     = '0.12.1'
PLUGIN_DESCRIPTION = _(r"""Sets the online status of your instant messenger client according to the music you are listening to.\n\nSupported IM clients are:\n * Gaim (>= 2.0 beta6)\n * Gajim\n * Gossip\n * Pidgin""")

# Possible actions when Exaile quits or stops playing
(
    DO_NOTHING,
    CHANGE_STATUS
) = range(2)

# Constants
PID                     = plugins.name(__file__)
TRACK_FIELDS            = ('album', 'artist', 'bitrate', 'genre', 'length', 'rating', 'title', 'track', 'year')
DEFAULT_STOP_ACTION     = CHANGE_STATUS
DEFAULT_STOP_STATUS     = _('Exaile is stopped')
# TRANSLATORS: IMStatus plugin default status format
DEFAULT_STATUS_FORMAT   = _('♫ {artist} - {album} ♫')
DEFAULT_UPDATE_ON_PAUSE = False

# Global variables
mHandlers           = plugins.SignalContainer()  # Handlers used by this plugin
mIMClients          = []                         # Active IM clients
mNowPlaying         = ''                         # The currently used status message
mCurrentTrack       = None                       # The track currently being played
mSupportedIMClients = []                         # Supported IM clients: this list is populated later in the code

##############################################################################

class Pidgin :

    def __init__(self, dbusInterface) :
        """
            Constructor
        """
        self.dbusInterface = dbusInterface

    def listAccounts(self) :
        """
            Purple merges all accounts, so we return a default one
            Each account is associated with:
                * A boolean -> True if the status of this account was changed on the previous track change
        """
        return {'GenericAccount' : False}

    def setAccountStatusMsg(self, account, msg) :
        """
            Change the status message of the given account
            Why is it so complex with Purple??
            Return true if the message is successfully updated
        """
        try :
            current    = self.dbusInterface.PurpleSavedstatusGetCurrent()
            statusType = self.dbusInterface.PurpleSavedstatusGetType(current)
            statusId   = self.dbusInterface.PurplePrimitiveGetIdFromType(statusType)
            if statusId in ['available'] :
                saved = self.dbusInterface.PurpleSavedstatusNew('', statusType)
                self.dbusInterface.PurpleSavedstatusSetMessage(saved, msg)
                self.dbusInterface.PurpleSavedstatusActivate(saved)
                return True
        except :
            pass

        return False

##############################################################################

class Gaim :

    def __init__(self, dbusInterface) :
        """
            Constructor
        """
        self.dbusInterface = dbusInterface

    def listAccounts(self) :
        """
            Gaim merges all accounts, so we return a default one
            Each account is associated with:
                * A boolean -> True if the status of this account was changed on the previous track change
        """
        return {'GenericAccount' : False}

    def setAccountStatusMsg(self, account, msg) :
        """
            Change the status message of the given account
            Why is it so complex with Gaim??
            Return true if the message is successfully updated
        """
        try :
            current    = self.dbusInterface.GaimSavedstatusGetCurrent()
            statusType = self.dbusInterface.GaimSavedstatusGetType(current)
            statusId   = self.dbusInterface.GaimPrimitiveGetIdFromType(statusType)
            if statusId in ['available'] :
                saved = self.dbusInterface.GaimSavedstatusNew('', statusType)
                self.dbusInterface.GaimSavedstatusSetMessage(saved, msg)
                self.dbusInterface.GaimSavedstatusActivate(saved)
                return True
        except :
            pass

        return False

##############################################################################

class Gajim :

    def __init__(self, dbusInterface) :
        """
            Constructor
        """
        self.dbusInterface = dbusInterface

    def listAccounts(self) :
        """
            Return a list of existing accounts
            Each account is associated with:
                * A boolean -> True if the status of this account was changed on the previous track change
        """
        try :
            return dict([(account, False) for account in self.dbusInterface.list_accounts()])
        except :
            return {}

    def setAccountStatusMsg(self, account, msg) :
        """
            Change the status message of the given account
            Return true if the message is successfully updated
        """
        try :
            currentStatus = self.dbusInterface.get_status(account)
            if currentStatus in ['online', 'chat'] :
                self.dbusInterface.change_status(currentStatus, msg, account)
                return True
        except :
            pass

        return False

##############################################################################

class Gossip :

    def __init__(self, dbusInterface) :
        """
            Constructor
        """
        self.dbusInterface = dbusInterface

    def listAccounts(self) :
        """
            Gossip merges all accounts, so we return a default one
            Each account is associated with:
                * A boolean -> True if the status of this account was changed on the previous track change
        """
        return {'GenericAccount' : False}

    def setAccountStatusMsg(self, account, msg) :
        """
            Change the status message of the given account
            Return true if the message is successfully updated
        """
        try :
            currentStatus, currentMsg = self.dbusInterface.GetPresence('')
            if currentStatus in ['available'] :
                self.dbusInterface.SetPresence(currentStatus, msg)
                return True
        except :
            pass

        return False

##############################################################################

# Elements associated with each supported IM clients
(
    IM_DBUS_SERVICE_NAME,
    IM_DBUS_OBJECT_NAME,
    IM_DBUS_INTERFACE_NAME,
    IM_CLASS,
    IM_INSTANCE,
    IM_ACCOUNTS
) = range(6)


# All specific classes have been defined, so we can now populate the list of supported IM clients
mSupportedIMClients = [
    ['im.pidgin.purple.PurpleService', '/im/pidgin/purple/PurpleObject',      'im.pidgin.purple.PurpleInterface',      Pidgin,   None, {}],  # Pidgin
    ['net.sf.gaim.GaimService',        '/net/sf/gaim/GaimObject',             'net.sf.gaim.GaimInterface',             Gaim,   None, {}],  # Gaim
    ['org.gajim.dbus',                 '/org/gajim/dbus/RemoteObject',        'org.gajim.dbus.RemoteInterface',        Gajim,  None, {}],  # Gajim
    ['org.gnome.Gossip',               '/org/gnome/Gossip',                   'org.gnome.Gossip',                      Gossip, None, {}]   # Gossip
]


def setStatusMsg(nowPlaying) :
    """
        Try to update the status of all accounts of all detected IM clients
    """
    global mNowPlaying

    for client in mIMClients :
        for account,wasUpdated in client[IM_ACCOUNTS].iteritems() :
            if nowPlaying != mNowPlaying or not wasUpdated :
                client[IM_ACCOUNTS][account] = client[IM_INSTANCE].setAccountStatusMsg(account, nowPlaying)
    mNowPlaying = nowPlaying


def updateStatus() :
    """
        Information about the current track has changed
    """
    global mCurrentTrack

    mCurrentTrack = APP.player.current

    # Construct the new status message
    nowPlaying = APP.settings.get_str('format', DEFAULT_STATUS_FORMAT, PID)
    for field in TRACK_FIELDS :
        nowPlaying = nowPlaying.replace('{%s}' % field, unicode(getattr(mCurrentTrack, field)))

    if APP.player.is_paused() and APP.settings.get_boolean('updateOnPause', DEFAULT_UPDATE_ON_PAUSE, PID) :
        nowPlaying = nowPlaying + ' [paused]'

    # And try to update the status of all IM clients
    setStatusMsg(nowPlaying)


def stop_track(exaile, track) :
    """
        Called when Exaile quits or when playback stops
    """
    global mCurrentTrack

    mCurrentTrack = None

    # Stop event is sent also when a track is finished, even if Exaile is going to play the next one
    # Since we don't want to change the status between each track, we add a delay of 1 sec. before doing anything
    gobject.timeout_add(1000, onStopTimer)


def onStopTimer() :
    """
        Called 1 sec. after the reception of a stop event
    """
    global mNowPlaying

    # If mCurrentTrack is still None, it means that Exaile has not started to play another track
    # In this case, we may assume that it is really stopped
    if mCurrentTrack is None :
        mNowPlaying = ''
        if APP.settings.get_int('stopAction', DEFAULT_STOP_ACTION, PID) == CHANGE_STATUS :
            setStatusMsg(APP.settings.get_str('stopStatus', DEFAULT_STOP_STATUS, PID))
    return False


def initialize(self = None):
    """
        Called when the plugin is enabled
    """
    global mIMClients
    oldNumber      = len(mIMClients) 
    mIMClients     = []
    activeServices = dbus.SessionBus().get_object('org.freedesktop.DBus', '/org/freedesktop/DBus').ListNames()

    for client in mSupportedIMClients :
        if client[IM_DBUS_SERVICE_NAME] in activeServices :
            obj       = dbus.SessionBus().get_object(client[IM_DBUS_SERVICE_NAME], client[IM_DBUS_OBJECT_NAME])
            interface = dbus.Interface(obj, client[IM_DBUS_INTERFACE_NAME])

            client[IM_INSTANCE] = client[IM_CLASS](interface)
            client[IM_ACCOUNTS] = client[IM_INSTANCE].listAccounts()

            mIMClients.append(client)
            
            
    # If there is at least one active IM client, we can connect our handlers
    if oldNumber == 0 :
        if len(mIMClients) != 0 :
            mHandlers.connect(APP.player, 'stop-track',                 stop_track)
            mHandlers.connect(APP.player, 'pause-toggled',              lambda exaile, track : updateStatus())
            mHandlers.connect(APP,        'track-information-updated',  lambda arg : updateStatus())
    else :
        if len(mIMClients) == 0 :
            mHandlers.disconnect_all()
    
    #print "Working clients: ", mIMClients

    return True

def destroy() :
    """
        Called when the plugin is disabled
    """
    mHandlers.disconnect_all()


def showHelp(widget, data=None) :
    """
        Display a dialog box with some help about status format
    """
    msg = _('Any field of the form <b>{field}</b> will be replaced by the '
        'corresponding value.\n\nAvailable fields are ')
    for field in TRACK_FIELDS :
        msg = msg + '<i>' + field + '</i>, '
    dlg = gtk.MessageDialog(None, gtk.DIALOG_MODAL, gtk.MESSAGE_INFO, gtk.BUTTONS_OK)
    dlg.set_markup(msg[:-2])
    dlg.run()
    dlg.destroy()


def configure() :
    """
        Called when the user wants to configure the plugin
    """
    dlg = plugins.PluginConfigDialog(APP.window, PLUGIN_NAME)

    align = gtk.Alignment()
    align.set_padding(5, 5, 5, 5)
    dlg.get_child().add(align)
    mainBox = gtk.VBox()
    mainBox.set_spacing(5)
    align.add(mainBox)

    # Upper part: status format
    frame = gtk.Frame('')
    frame.get_label_widget().set_markup(_('<b> Status message: </b>'))
    mainBox.pack_start(frame, True, True, 0)
    align = gtk.Alignment(0, 0, 1, 1)
    align.set_padding(5, 5, 5, 5)
    frame.add(align)
    tmpBox = gtk.HBox()
    tmpBox.set_spacing(5)
    entryFormat = gtk.Entry()
    entryFormat.set_text(APP.settings.get_str('format', DEFAULT_STATUS_FORMAT, PID))
    button = gtk.Button(_('Help'), gtk.STOCK_HELP)
    tmpBox.pack_start(entryFormat, True, True)
    tmpBox.pack_start(button)
    button.connect('clicked', showHelp)
    align.add(tmpBox)

    # Middle part: what should be done when Exaile quits of stops playing
    frame = gtk.Frame('')
    frame.get_label_widget().set_markup(_('<b> When Exaile stops playing or quits: </b>'))
    mainBox.pack_start(frame, False, False, 0)
    tmpBox = gtk.VBox()
    tmpBox.set_spacing(5)
    align = gtk.Alignment()
    align.set_padding(5, 5, 5, 5)
    frame.add(align)
    align.add(tmpBox)
    radioDoNothing = gtk.RadioButton(None, _('Do nothing'))
    radioSetStatus = gtk.RadioButton(radioDoNothing, _('Set status to:'))
    if APP.settings.get_int('stopAction', DEFAULT_STOP_ACTION, PID) == DO_NOTHING :
        radioDoNothing.set_active(True)
    else :
        radioSetStatus.set_active(True)
    tmpBox.pack_start(radioDoNothing, False, False, 0)
    box2 = gtk.HBox()
    entryStopStatus = gtk.Entry()
    entryStopStatus.set_text(APP.settings.get_str('stopStatus', DEFAULT_STOP_STATUS, PID))
    box2.pack_start(radioSetStatus, False, False, 0)
    box2.pack_start(entryStopStatus, False, False, 0)
    tmpBox.pack_start(box2, False, False, 0)

    # Lower part: miscellaneous
    frame = gtk.Frame('')
    frame.get_label_widget().set_markup(_('<b> Miscellaneous: </b>'))
    mainBox.pack_start(frame, False, False, 0)
    tmpBox = gtk.VBox()
    tmpBox.set_spacing(5)
    align = gtk.Alignment()
    align.set_padding(5, 5, 5, 5)
    frame.add(align)
    align.add(tmpBox)
    checkUpdateOnPause = gtk.CheckButton(_('Update status when track is paused'))
    tmpBox.pack_start(checkUpdateOnPause, False, False, 0)
    checkUpdateOnPause.set_active(APP.settings.get_boolean('updateOnPause', DEFAULT_UPDATE_ON_PAUSE, PID))
    resetButton = gtk.Button(_('Reset'))
    tmpBox.pack_start(resetButton, False, False, 0)
    resetButton.connect('clicked', initialize)

    dlg.show_all()
    result = dlg.run()
    dlg.hide()

    if result == gtk.RESPONSE_OK :
        if radioDoNothing.get_active() :
            APP.settings.set_int('stopAction', DO_NOTHING, PID)
        else :
            APP.settings.set_int('stopAction', CHANGE_STATUS, PID)
        APP.settings.set_str('stopStatus',        entryStopStatus.get_text(),      PID)
        APP.settings.set_str('format',            entryFormat.get_text(),          PID)
        APP.settings.set_boolean('updateOnPause', checkUpdateOnPause.get_active(), PID)
        if mCurrentTrack is not None :
            updateStatus()



#!/usr/bin/env python

import os, re

h = os.popen('ps x | grep exaile')
lines = h.readlines()

for line in lines:
    info = re.split(r'\s+', 
        line.strip())
    if len(info) == 7:
        if info[4] == 'python' and info[6] \
            == 'exaile.py':
            os.system('kill -9 %s' % info[0])
